# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Module for building position mapping models based on a flatmap"""

import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import progressbar
from scipy.interpolate import griddata
from scipy.spatial import distance_matrix
from voxcell.nexus.voxelbrain import Atlas

from connectome_manipulator import log
from connectome_manipulator.model_building import model_types


def extract(
    circuit,
    flatmap_path,
    xy_file,
    z_file,
    xy_scale=None,
    z_scale=None,
    nodes_pop_name=None,
    NN_only=False,
    CV_dict=None,
    **_,
):
    """Extracts a position mapping from 3D atlas space (x/y/z) to 3D flat space (flat-x/flat-y/depth) of a given population of neurons.

    Args:
        circuit (bluepysnap.Circuit): Input circuit
        flatmap_path (str): Base path to a flatmap
        xy_file (str): Filename of x/y mapping file as part of the flatmap
        z_file (str): Filename of z (= cortical depth) mapping file as part of the flatmap
        xy_scale (list-like): Two-element list with x/y scaling factors from a.u. (as in flatmap) to um
        z_scale (float): Scalar value with z scaling factor from a.u. (as in flatmap) to um
        nodes_pop_name (str): Name of SONATA nodes population to extract data from
        NN_only (bool): If selected, only nearest-neighbor interpolation will be used for position mapping (faster); otherwise, linear interpolation is applied, if possible (slower)
        CV_dict (dict): Cross-validation dictionary - Not supported

    Returns:
        dict: Dictionary containing the extracted data elements, i.e., neuron positions in original and flat space
    """
    log.log_assert(CV_dict is None, "ERROR: Cross-validation not supported!")

    # Get neuron positions
    if nodes_pop_name is None:
        log.log_assert(
            len(circuit.nodes.population_names) == 1,
            f"ERROR: Nodes population could not be determined (found {circuit.nodes.population_names})!",
        )
        nodes_pop_name = circuit.nodes.population_names[0]
        log.debug(f'Loading nodes population "{nodes_pop_name}"')
    nodes = circuit.nodes[nodes_pop_name]
    nrn_pos = nodes.positions()
    nrn_ids = nrn_pos.index.to_numpy()
    nrn_pos = nrn_pos.to_numpy()
    nrn_lay = nodes.get(properties="layer")

    # Load flatmap
    flatmap_atlas = Atlas.open(flatmap_path)
    flatmap = flatmap_atlas.load_data(xy_file)
    depths = flatmap_atlas.load_data(z_file)
    log.log_assert(
        flatmap.raw.shape[:3] == depths.raw.shape,
        "ERROR: Flatmap and depths map sizes inconsistent!",
    )
    log.log_assert(
        np.all(flatmap.voxel_dimensions == depths.voxel_dimensions),
        "ERROR: Flatmap and depths map voxels inconsistent!",
    )
    log.log_assert(
        np.all(flatmap.bbox == depths.bbox),
        "ERROR: Flatmap and depths map bounding boxes inconsistent!",
    )

    if xy_scale is None:  # x/y scaling from a.u. to um
        xy_scale = [
            flatmap.voxel_dimensions[0],
            flatmap.voxel_dimensions[1],
        ]  # Default: Assume same pixel size in flat space as voxel size in atlas
    else:
        log.log_assert(
            np.array(xy_scale).size == 2
            and np.all(np.isfinite(xy_scale))
            and np.all(np.array(xy_scale) != 0),
            "ERROR: XY scale error!",
        )

    if z_scale is None:  # z scaling from a.u. to um
        z_scale = 1.0  # Default: Assume depth values are already scaled to um
    else:
        log.log_assert(
            np.array(z_scale).size == 1 and np.isfinite(z_scale) and z_scale != 0,
            "ERROR: Z scale error!",
        )

    log.info(
        f'Loaded x/y flatmap ("{xy_file}"; scale={np.round(xy_scale, decimals=2)}) and z (depth) map ("{z_file}"; scale={np.round(z_scale, decimals=2)}) from "{flatmap_path}"'
    )

    # Convert cell positions to flat space [Assume: missing values set to -1]
    flat_x = flatmap.lookup(nrn_pos)[:, 0].astype(float)
    flat_x[flat_x != -1] = flat_x[flat_x != -1] * xy_scale[0]
    flat_y = flatmap.lookup(nrn_pos)[:, 1].astype(float)
    flat_y[flat_y != -1] = flat_y[flat_y != -1] * xy_scale[1]
    flat_z = depths.lookup(nrn_pos).astype(float)
    flat_z[flat_z != -1] = flat_z[flat_z != -1] * z_scale

    # Determine map indices/positions
    map_indices = flatmap.positions_to_indices(
        nrn_pos, keep_fraction=True
    )  # Keep fractions within voxels => fraction x.5 corresponds to voxel center
    map_pos = np.floor(map_indices) + 0.5  # Voxel values assumed to correspond to voxel center

    # Nearest-neighbor interpolation only (FASTER)
    if NN_only:
        log.debug("Using nearest-neighbor interpolation only!")
        flat_x_intpl = griddata(
            map_pos[flat_x != -1], flat_x[flat_x != -1], map_indices, method="nearest"
        )
        flat_y_intpl = griddata(
            map_pos[flat_y != -1], flat_y[flat_y != -1], map_indices, method="nearest"
        )
        flat_z_intpl = griddata(
            map_pos[flat_z != -1], flat_z[flat_z != -1], map_indices, method="nearest"
        )
    else:
        # Linear interpolation, if possible
        flat_x_intpl = griddata(map_pos[flat_x != -1], flat_x[flat_x != -1], map_indices)
        flat_y_intpl = griddata(map_pos[flat_y != -1], flat_y[flat_y != -1], map_indices)
        flat_z_intpl = griddata(map_pos[flat_z != -1], flat_z[flat_z != -1], map_indices)

        # Nearest-neighbor interpolation, otherwise
        flat_x_intpl[np.isnan(flat_x_intpl)] = griddata(
            map_pos[flat_x != -1],
            flat_x[flat_x != -1],
            map_indices[np.isnan(flat_x_intpl)],
            method="nearest",
        )
        flat_y_intpl[np.isnan(flat_y_intpl)] = griddata(
            map_pos[flat_y != -1],
            flat_y[flat_y != -1],
            map_indices[np.isnan(flat_y_intpl)],
            method="nearest",
        )
        flat_z_intpl[np.isnan(flat_z_intpl)] = griddata(
            map_pos[flat_z != -1],
            flat_z[flat_z != -1],
            map_indices[np.isnan(flat_z_intpl)],
            method="nearest",
        )

    flat_pos = np.vstack((flat_x_intpl, flat_y_intpl, flat_z_intpl)).T

    return {"nrn_ids": nrn_ids, "nrn_lay": nrn_lay, "nrn_pos": nrn_pos, "flat_pos": flat_pos}


def build(nrn_ids, flat_pos, **_):
    """Builds a flat space position mapping model from data.

    Args:
        nrn_ids (list-like): List of mapped neuron IDs, as returned by :func:`extract`
        flat_pos (numpy.ndarray): Table of mapped neuron positions in 3D flat space of size <#neurons x 3>, as returned by :func:`extract`

    Returns:
        connectome_manipulator.model_building.model_types.PosMapModel: Resulting position mapping model
    """
    flat_pos_table = pd.DataFrame(flat_pos, index=nrn_ids, columns=["x", "y", "z"])
    log.log_assert(np.all(np.isfinite(flat_pos_table)), "ERROR: Position error!")

    # Create model
    model = model_types.PosMapModel(pos_table=flat_pos_table)
    log.debug("Model description:\n%s", model)

    return model


def plot(out_dir, nrn_ids, nrn_lay, nrn_pos, model, **_):  # pragma: no cover
    """Visualizes neuron positions in original space vs. mapped space from model output.

    Args:
        out_dir (str): Path to output directory where the results figures will be stored
        nrn_ids (list-like): List of mapped neuron IDs, as returned by :func:`extract`
        nrn_lay (list-like): List of layer property values for all mapped neurons, as returned by :func:`extract`
        nrn_pos (numpy.ndarray): Table of original neuron positions in 3D atlas space of size <#neurons x 3>, as returned by :func:`extract`
        model (connectome_manipulator.model_building.model_types.PosMapModel): Resulting position mapping model, as returned by :func:`build`
    """
    nrn_pos_model = model.apply(gids=nrn_ids)

    # 3D cell positions in atlas vs. flat space
    num_layers = len(np.unique(nrn_lay))
    lay_colors = plt.get_cmap("jet")(np.linspace(0, 1, num_layers))
    views = [[90, 0], [0, 0]]
    pos_list = [nrn_pos, nrn_pos_model]
    lbl_list = ["Atlas space (data)", "Flat space (model)"]
    fig = plt.figure(figsize=(10, 3 * len(views)), dpi=300)
    plt.gcf().patch.set_facecolor("w")
    for vidx, v in enumerate(views):
        for pidx, (pos, lbl) in enumerate(zip(pos_list, lbl_list)):
            ax = fig.add_subplot(
                len(views), len(pos_list), vidx * len(pos_list) + pidx + 1, projection="3d"
            )
            for lidx in range(num_layers):
                pos_sel = pos[nrn_lay == lidx + 1, :]
                plt.plot(
                    pos_sel[:, 0],
                    pos_sel[:, 1],
                    pos_sel[:, 2],
                    ".",
                    color=lay_colors[lidx, :],
                    markersize=1.0,
                    alpha=0.5,
                    label=f"L{lidx + 1}",
                )
            ax.view_init(*v)
            ax.set_xlabel("x [$\\mu$m]")
            ax.set_ylabel("y [$\\mu$m]")
            ax.set_zlabel("z [$\\mu$m]")
            plt.legend(loc="center left", bbox_to_anchor=(1.0, 0.5), ncol=1)
            if vidx == 0:
                plt.title(lbl + f"\n[N={len(nrn_ids)}cells]")
    plt.tight_layout()

    out_fn = os.path.abspath(os.path.join(out_dir, "data_vs_model_positions.png"))
    log.info(f"Saving {out_fn}...")
    plt.savefig(out_fn)

    # Cell distances in atlas vs. flat space
    max_plot = 10000
    if len(nrn_ids) > max_plot:
        log.debug("Using subsampling for distance plots!")
        nrn_sel = np.random.choice(len(nrn_ids), max_plot)
        nrn_ids = nrn_ids[nrn_sel]
        nrn_pos = nrn_pos[nrn_sel, :]
        nrn_pos_model = nrn_pos_model[nrn_sel, :]

    dist_mat_data = distance_matrix(nrn_pos, nrn_pos)
    dist_mat_model = distance_matrix(nrn_pos_model, nrn_pos_model)

    triu_idx = np.triu_indices(len(nrn_ids), 1)
    dist_val_data = dist_mat_data[triu_idx]
    dist_val_model = dist_mat_model[triu_idx]

    dist_max = max(*dist_val_data, *dist_val_model)
    plt.figure(figsize=(5, 5), dpi=300)
    plt.plot(dist_val_data, dist_val_model, "b.", alpha=0.1, markersize=1.0, markeredgecolor="none")
    plt.plot([0, dist_max], [0, dist_max], "k--")
    plt.xlim((0, dist_max))
    plt.ylim((0, dist_max))
    plt.grid(True)
    plt.xlabel("Distance in atlas space (data) [$\\mu$m]")
    plt.ylabel("Distance in flat space (model) [$\\mu$m]")
    plt.title(f"Cell distances in atlas vs. flat space [N={len(nrn_ids)}cells]")
    plt.tight_layout()

    out_fn = os.path.abspath(os.path.join(out_dir, "data_vs_model_distances.png"))
    log.info(f"Saving {out_fn}...")
    plt.savefig(out_fn)

    # Nearest neighbors in atlas vs. flat space
    NN_mat_data = np.argsort(dist_mat_data, axis=1)
    NN_mat_model = np.argsort(dist_mat_model, axis=1)

    num_NN_list = list(range(1, 30, 1))
    NN_match = np.full(len(num_NN_list), np.nan)

    log.debug("Computing nearest neighbors in atlas vs. flat space...")
    pbar = progressbar.ProgressBar()
    for nidx in pbar(range(len(num_NN_list))):
        num_NN = num_NN_list[nidx]
        NN_match[nidx] = np.mean(
            [
                len(np.intersect1d(NN_mat_data[i, 1 : 1 + num_NN], NN_mat_model[i, 1 : 1 + num_NN]))
                / num_NN
                for i in range(len(nrn_ids))
            ]
        )

    plt.figure(figsize=(5, 4), dpi=300)
    plt.plot(num_NN_list, NN_match, ".-")
    plt.grid(True)
    plt.ylim((0, 1))
    plt.xlabel("#Nearest neighbors")
    plt.ylabel("Mean match")
    plt.title(f"Nearest neighbors in atlas vs. flat space [N={len(nrn_ids)}cells]")

    out_fn = os.path.abspath(os.path.join(out_dir, "data_vs_model_neighbors.png"))
    log.info(f"Saving {out_fn}...")
    plt.savefig(out_fn)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import json
import numpy as np
import os
import pytest

from bluepysnap import Circuit

from utils import TEST_DATA_DIR
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation
from connectome_manipulator.connectome_manipulation.converters import EdgeWriter
from connectome_manipulator import log


@pytest.fixture
def manipulation():
    m = Manipulation.get("conn_extraction")
    return m


def test_apply(manipulation):
    log.setup_logging()  # To have data logging in a defined state

    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = circuit.edges[circuit.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)

    # Test that given intrinsic cell target is extracted (no extra node sets file given)
    for tgt_name in ["LayerA", "RegionB"]:
        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(tgt_ids, target_name=tgt_name, node_sets_file=None)
        res = writer.to_pandas()
        src_ids = nodes[0].ids(tgt_name)
        tgt_ids = nodes[1].ids(tgt_name)
        assert np.all(np.isin(res["@source_node"], src_ids)) and np.all(
            np.isin(res["@target_node"], tgt_ids)
        )
        assert (
            np.sum(
                np.logical_and(
                    np.isin(edges_table["@source_node"], src_ids),
                    np.isin(edges_table["@target_node"], tgt_ids),
                )
            )
            == res.shape[0]
        )

    # Test that given external cell target is extracted (given by extra node sets file)
    node_sets_file = os.path.join(TEST_DATA_DIR, "node_sets_extra.json")
    with open(node_sets_file, "r") as f:
        node_sets = json.load(f)
    for tgt_name in ["NSet1", "NSet1"]:
        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids, target_name=tgt_name, node_sets_file=node_sets_file
        )
        res = writer.to_pandas()
        src_ids = node_sets[tgt_name]["node_id"]
        tgt_ids = node_sets[tgt_name]["node_id"]
        assert np.all(np.isin(res["@source_node"], src_ids)) and np.all(
            np.isin(res["@target_node"], tgt_ids)
        )
        assert (
            np.sum(
                np.logical_and(
                    np.isin(edges_table["@source_node"], src_ids),
                    np.isin(edges_table["@target_node"], tgt_ids),
                )
            )
            == res.shape[0]
        )

    # Test special case when no cell target specified (should return empty connectome)
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(tgt_ids, target_name=None, node_sets_file=None)
    res = writer.to_pandas()
    assert res.empty
    assert np.all(res.columns == edges_table.columns)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Test that alternate morphologies work."""

import os
import pytest
import bluepysnap
from connectome_manipulator.connectome_manipulation.converters import EdgeWriter
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation
from utils import TEST_DATA_DIR


@pytest.fixture
def manipulation():
    m = Manipulation.get("conn_wiring")
    return m


def test_apply(manipulation):
    c = bluepysnap.Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata_h5.json"))
    edges = c.edges[c.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)
    edges_table_empty = edges_table.loc[[]].copy()
    writer = EdgeWriter(None, edges_table_empty)
    m = manipulation(nodes, writer)
    with pytest.raises(bluepysnap.exceptions.BluepySnapError):
        m.morpho_helper.get(tgt_ids[0], extension="swc")
    m.morpho_helper.get(tgt_ids[0], extension="h5")



# This file is part of connectome-manipulator.

"""Module for comparing connectomes based on connectivity features:

Structural comparison of two connectomes in terms of connection probability matrices for selected
pathways (including #synapses per connection), as specified by the config. For each connectome,
the underlying connectivity matrices are computed by the :func:`compute` function and will be saved
to a data file first. The individual connectivity matrices, together with a difference map
between the two connectomes, are then plotted by means of the :func:`plot` function.
"""

import matplotlib.pyplot as plt
import numpy as np
import progressbar
import pandas as pd

from scipy.spatial import KDTree
from scipy import sparse

from connectome_manipulator.access_functions import (
    get_edges_population,
    get_node_ids,
    get_connections,
)


def within_max_distance_matrix(pre_neurons, post_neurons, max_dist, props_for_distance):
    """Computes a sparse bool matrix of neuron pairs within a specified maximum distance. The value of the matrix at i, j is True iff the pair of neuron i and neuron j are within that distance.

    Args:
        pre_neurons (tuple): A tuple of the node population object and the node ids for the first population of neurons. This population will be indexed along the first axis of the output
        post_neurons (tuple): A tuple of the node population object and the node ids for the second population of neurons. This population will be indexed along the second axis of the output
        max_dist (float): Maximum distance to use.
        props_for_distance (list): List of node properties that must be available for both populations. Their values must be numeric. They will be used to calculate the (Euclidean) distance.
    """
    nodes_pre, pre_ids = pre_neurons
    nodes_post, post_ids = post_neurons
    assert np.all([_p in nodes_pre.property_dtypes for _p in props_for_distance])
    assert np.all([_p in nodes_post.property_dtypes for _p in props_for_distance])
    if len(pre_ids) == 0 or len(post_ids) == 0:
        within_mat = sparse.csr_matrix((len(pre_ids), len(post_ids)), dtype=bool)
        return within_mat, pd.Series([]), pd.Series([])

    locs_pre = nodes_pre.get(pre_ids, props_for_distance)
    locs_post = nodes_post.get(post_ids, props_for_distance)

    lookup_pre = pd.Series(
        range(len(locs_pre)), index=locs_pre.index
    )  # from node id to 0, 1, 2, ...
    lookup_post = pd.Series(range(len(locs_post)), index=locs_post.index)

    tree_pre = KDTree(locs_pre)
    tree_post = KDTree(locs_post)

    pairs_within = tree_pre.query_ball_tree(tree_post, max_dist)
    indptr = np.cumsum([0] + list(map(len, pairs_within)))
    indices = np.hstack(pairs_within)
    within_mat = sparse.csr_matrix(
        (np.ones_like(indices, dtype=bool), indices, indptr), shape=(len(locs_pre), len(locs_post))
    )
    return within_mat, lookup_pre, lookup_post


def compute(
    circuit,
    group_by=None,
    sel_src=None,
    sel_dest=None,
    skip_empty_groups=False,
    edges_popul_name=None,
    max_distance=None,
    props_for_distance=None,
    **_,
):
    """Computes the average connection probabilities and #synapses/connection between groups of neurons of a given circuit's connectome.

    Args:
        circuit (bluepysnap.Circuit): Input circuit
        group_by (str): Neuron property name based on which to group connections, e.g., "synapse_class", "layer", or "mtype"; if omitted, the overall average is computed
        sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
        sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
        skip_empty_groups (bool): If selected, only group property values that exist within the given source/target selection are kept; otherwise, all group property values, even if not present in the given source/target selection, will be included
        edges_popul_name (str): Name of SONATA egdes population to extract data from
        max_distance (float): Optional. Maximum distance of pairs of neurons considered. If used, must also provide ``props_for_distance``.
        props_for_distance (list): Optional. To be provided with ``max_distance``. Numerical node properties that are used to calculate the pairwise distances.

    Returns:
        dict: Dictionary containing the computed data elements; see Notes

    Note:
        The returned dictionary contains the following data elements that can be selected for plotting through the structural comparison configuration file, together with a common dictionary containing additional information. Each data element is a dictionary with "data" (numpy.ndarray of size <source-group-size x target-group-size>), "name" (str), and "unit" (str) items.

        * "nsyn_conn": Mean number of synapses per connection
        * "nsyn_conn_std": Standard deviation of the number of synapses per connection
        * "nsyn_conn_sem": Standard error of the mean of the number of synapses per connection
        * "nsyn_conn_min": Minimum number of synapses per connection
        * "nsyn_conn_max": Maximum number of synapses per connection
        * "conn_prob": Average connection probability
    """
    if max_distance is not None:
        assert (
            props_for_distance is not None
        ), "When specifying distance cutoff, must also specify properties to use!"
    # Select edge population
    edges = get_edges_population(circuit, edges_popul_name)

    # Select corresponding source/target nodes populations
    src_nodes = edges.source
    tgt_nodes = edges.target

    if group_by is None:
        src_group_sel = [sel_src]
        tgt_group_sel = [sel_dest]
    else:
        if (
            skip_empty_groups
        ):  # Take only group property values that exist within given src/tgt selection
            src_group_values = np.unique(
                src_nodes.get(get_node_ids(src_nodes, sel_src), properties=group_by)
            )
            tgt_group_values = np.unique(
                tgt_nodes.get(get_node_ids(tgt_nodes, sel_dest), properties=group_by)
            )
        else:  # Keep all group property values, even if not present in given src/tgt selection, to get the full matrix
            src_group_values = sorted(src_nodes.property_values(group_by))
            tgt_group_values = sorted(tgt_nodes.property_values(group_by))

        if sel_src is None:
            sel_src = {}
        else:
            assert isinstance(
                sel_src, dict
            ), "ERROR: Source node selection must be a dict or empty!"  # Otherwise, it cannot be merged with group selection
        if sel_dest is None:
            sel_dest = {}
        else:
            assert isinstance(
                sel_dest, dict
            ), "ERROR: Target node selection must be a dict or empty!"  # Otherwise, it cannot be merged with group selection

        src_group_sel = [
            {**sel_src, group_by: src_group_values[idx]} for idx in range(len(src_group_values))
        ]  # group_by will overwrite selection in case group property also exists in selection!
        tgt_group_sel = [
            {**sel_dest, group_by: tgt_group_values[idx]} for idx in range(len(tgt_group_values))
        ]  # group_by will overwrite selection in case group property also exists in selection!

    print(
        f"INFO: Computing connectivity (group_by={group_by}, sel_src={sel_src}, sel_dest={sel_dest}, N={len(src_group_values)}x{len(tgt_group_values)} groups, max_distance={max_distance} based on {props_for_distance})",
        flush=True,
    )

    syn_table = np.zeros((len(src_group_sel), len(tgt_group_sel)))  # Mean
    syn_table_std = np.zeros((len(src_group_sel), len(tgt_group_sel)))  # Std
    syn_table_sem = np.zeros((len(src_group_sel), len(tgt_group_sel)))  # SEM
    syn_table_min = np.zeros((len(src_group_sel), len(tgt_group_sel)))  # Min
    syn_table_max = np.zeros((len(src_group_sel), len(tgt_group_sel)))  # Max
    p_table = np.zeros((len(src_group_sel), len(tgt_group_sel)))
    pbar = progressbar.ProgressBar()
    for idx_pre in pbar(range(len(src_group_sel))):
        sel_pre = src_group_sel[idx_pre]
        for idx_post, _ in enumerate(tgt_group_sel):
            sel_post = tgt_group_sel[idx_post]
            pre_ids = get_node_ids(src_nodes, sel_pre)
            post_ids = get_node_ids(tgt_nodes, sel_post)
            conns = get_connections(edges, pre_ids, post_ids, with_nsyn=True)
            npairs = len(pre_ids) * len(post_ids)

            if conns.size > 0:
                if max_distance is not None and len(pre_ids) > 0 and len(post_ids) > 0:
                    M, lo_pre, lo_post = within_max_distance_matrix(
                        (src_nodes, pre_ids),
                        (tgt_nodes, post_ids),
                        max_distance,
                        props_for_distance,
                    )
                    is_within = np.asarray(M[lo_pre[conns[:, 0]], lo_post[conns[:, 1]]]).flatten()
                    conns = conns[is_within]
                    npairs = M.nnz

            if conns.size > 0:
                scounts = conns[:, 2]  # Synapse counts per connection
                ccount = len(scounts)  # Connection count

                syn_table[idx_pre, idx_post] = np.mean(scounts)
                syn_table_std[idx_pre, idx_post] = np.std(scounts)
                syn_table_sem[idx_pre, idx_post] = np.std(scounts) / np.sqrt(ccount)
                syn_table_min[idx_pre, idx_post] = np.min(scounts)
                syn_table_max[idx_pre, idx_post] = np.max(scounts)
                p_table[idx_pre, idx_post] = 100.0 * ccount / npairs

    syn_table_name = "Synapses per connection"
    syn_table_unit = "#syn/conn"
    p_table_name = "Connection probability"
    p_table_unit = "Conn. prob. (%)"

    return {
        "nsyn_conn": {"data": syn_table, "name": syn_table_name, "unit": "Mean " + syn_table_unit},
        "nsyn_conn_std": {
            "data": syn_table_std,
            "name": syn_table_name,
            "unit": "Std of " + syn_table_unit,
        },
        "nsyn_conn_sem": {
            "data": syn_table_sem,
            "name": syn_table_name,
            "unit": "SEM of " + syn_table_unit,
        },
        "nsyn_conn_min": {
            "data": syn_table_min,
            "name": syn_table_name,
            "unit": "Min " + syn_table_unit,
        },
        "nsyn_conn_max": {
            "data": syn_table_max,
            "name": syn_table_name,
            "unit": "Max " + syn_table_unit,
        },
        "conn_prob": {"data": p_table, "name": p_table_name, "unit": p_table_unit},
        "common": {"src_group_values": src_group_values, "tgt_group_values": tgt_group_values},
    }


def plot(
    res_dict, common_dict, fig_title=None, vmin=None, vmax=None, isdiff=False, group_by=None, **_
):  # pragma:no cover
    """Plots a connectivity matrix or a difference matrix.

    Args:
        res_dict (dict): Results dictionary, containing selected data for plotting; must contain a "data" item with a connectivity matrix of type numpy.ndarray of size  <#source-group-values x #target-group-values>, as well as "name" and "unit" items containing strings.
        common_dict (dict): Common dictionary, containing additional information; must contain "src_group_values" and "tgt_group_values" items containing lists of source/target values of the grouped property, matching the size of the connectivity matrix in ``res_dict``
        fig_title (str): Optional figure title
        vmin (float): Minimum plot range
        vmax (float): Maximum plot range
        isdiff (bool): Flag indicating that ``res_dict`` contains a difference matrix; in this case, a symmetric plot range is required and a divergent colormap will be used
        group_by (str): Neuron property name based on which to group connections, e.g., "synapse_class", "layer", or "mtype"; if omitted, the overall average is computed
    """
    if isdiff:  # Difference plot
        assert -1 * vmin == vmax, "ERROR: Symmetric plot range required!"
        cmap = "PiYG"  # Symmetric (diverging) colormap
    else:  # Regular plot
        cmap = "hot_r"  # Regular colormap

    plt.imshow(res_dict["data"], interpolation="nearest", cmap=cmap, vmin=vmin, vmax=vmax)

    if fig_title is None:
        plt.title(res_dict["name"])
    else:
        plt.title(fig_title)

    if group_by:
        plt.xlabel(f"Postsynaptic {group_by}")
        plt.ylabel(f"Presynaptic {group_by}")

    n_grp = np.maximum(len(common_dict["src_group_values"]), len(common_dict["tgt_group_values"]))
    font_size = max(13 - n_grp / 6, 1)  # Font scaling
    if len(common_dict["src_group_values"]) > 0:
        plt.yticks(
            range(len(common_dict["src_group_values"])),
            common_dict["src_group_values"],
            rotation=0,
            fontsize=font_size,
        )

    if len(common_dict["tgt_group_values"]) > 0:
        if max(len(str(grp)) for grp in common_dict["tgt_group_values"]) > 1:
            rot_x = 90
        else:
            rot_x = 0
        plt.xticks(
            range(len(common_dict["tgt_group_values"])),
            common_dict["tgt_group_values"],
            rotation=rot_x,
            fontsize=font_size,
        )

    cb = plt.colorbar()
    cb.set_label(res_dict["unit"])



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import glob
import json
import os

from mock import patch, Mock
from pathlib import Path
import pytest

from bluepysnap import Circuit
from bluepysnap.nodes import NodePopulation, Nodes
from bluepysnap.edges import EdgePopulation

from utils import setup_tempdir, TEST_DATA_DIR

import connectome_manipulator.connectome_manipulation.connectome_manipulation as test_module
import connectome_manipulator.connectome_manipulation.converters as converters
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation


def test_load_circuit():
    config_path = os.path.join(TEST_DATA_DIR, "circuit_sonata.json")

    # Check error handling if no edge population exists
    with patch(
        f"connectome_manipulator.connectome_manipulation.connectome_manipulation.Circuit.edges"
    ) as patched:
        patched.return_value = Mock(population_names=[])
        # patched.return_value = Mock(edges=Mock(population_names=[]))

        # UPDATE 10/11/2022: Circuits w/o edge population can be loaded and won't rise an error any more.
        #                    Returned edges-related values will be None.
        # expected_error = 'No edge population found!'
        # with pytest.raises(AssertionError, match=expected_error):
        #     test_module.load_circuit('fake_config')
        _, *res = test_module.load_circuit(config_path)

        # Check that there are 2 instances of NodePopulation in first item
        assert len(res[0]) == 2
        assert all(isinstance(r, NodePopulation) for r in res[0])

        # Check that two paths are returned and they're correct
        assert len(res[1]) == 2
        assert all(r == os.path.join(TEST_DATA_DIR, "nodes.h5") for r in res[1])

        # Check that edges-related values are None
        assert res[3] is res[4] is None

    # Check error handling if more than one non-default populations exist and no population name is specified
    with patch(
        f"connectome_manipulator.connectome_manipulation.connectome_manipulation.Circuit"
    ) as patched:
        patched.return_value = Mock(edges=Mock(population_names=["name1", "name2"]))

        expected_error = 'Population "default" not found in edges file!'
        with pytest.raises(AssertionError, match=expected_error):
            test_module.load_circuit("fake_config")

    res = test_module.load_circuit(config_path)

    # Check valid config
    config = res[0]
    nodes = config["networks"]["nodes"][0]
    nodes_file = nodes["nodes_file"]
    assert nodes_file == str(Path(TEST_DATA_DIR, "nodes.h5"))

    edges = config["networks"]["edges"][0]
    edges_file = edges["edges_file"]
    assert edges_file == str(Path(TEST_DATA_DIR, "edges.h5"))

    # Check that there are 2 instances of NodePopulation in first item
    assert len(res[1]) == 2
    assert all(isinstance(r, NodePopulation) for r in res[1])

    # Check that two paths are returned and they're correct
    assert len(res[2]) == 2
    assert all(r == os.path.join(TEST_DATA_DIR, "nodes.h5") for r in res[2])

    # Check that last edge population is returned
    assert isinstance(res[3], EdgePopulation)

    # Check that a correct edge path is returned
    assert res[4] == os.path.join(TEST_DATA_DIR, "edges.h5")


@patch("connectome_manipulator.connectome_manipulation.manipulation.base.get_enumeration_map")
def test_manipulation_registration(nodes):
    manip_mods = glob.glob("connectome_manipulator/connectome_manipulation/manipulation/*.py")
    manip_mods = [os.path.splitext(os.path.basename(m))[0] for m in manip_mods]
    print(manip_mods)
    for mod_name in manip_mods:
        if mod_name not in ["base", "__init__"]:
            m = Manipulation.get(mod_name)(nodes, None)


@patch("connectome_manipulator.connectome_manipulation.manipulation.base.get_enumeration_map")
def test_null_manipulation(nodes):
    module_name = "null_manipulation"
    manip_config = {"manip": {"name": "test", "fcts": [{"source": module_name, "kwargs": {}}]}}

    m = Manipulation.get(module_name)

    writer = converters.EdgeWriter(None)
    m(nodes, writer).apply(None, **manip_config)
    assert len(writer.to_pandas()) == 0


@patch("connectome_manipulator.connectome_manipulation.manipulation.base.get_enumeration_map")
def test_null_manipulation_with_edges(nodes):
    module_name = "null_manipulation"
    manip_config = {"manip": {"name": "test", "fcts": [{"source": module_name, "kwargs": {}}]}}

    m = Manipulation.get(module_name)

    c = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = c.edges[c.edges.population_names[0]]
    tgt_ids = edges.target.ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)

    writer = converters.EdgeWriter(None, edges_table.copy())
    m(nodes, writer).apply(None, **manip_config)
    assert edges_table.equals(writer.to_pandas())


def test_edges_to_parquet(tmp_path):
    c = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = c.edges[c.edges.population_names[0]]
    edges_table = edges.afferent_edges([0, 1], properties=sorted(edges.property_names))

    outfile = tmp_path / "test.parquet"
    with converters.EdgeWriter(outfile, existing_edges=edges_table) as writer:
        pass
    assert outfile.is_file()


def test_parquet_to_sonata():
    class FakeNode:
        name = ""

    infiles = ["dummy"]
    inpath = "some/path"
    nodes = [FakeNode(), FakeNode()]
    nodefiles = ["fake/path", "another/one"]

    class MockGlob:
        def __init__(self, *args, **kwargs):
            pass

        @staticmethod
        def glob(path):
            return infiles

    class MockPopen:
        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs):
            pass

        @staticmethod
        def communicate():
            return ["".encode()]

    class MockReadMetadataEmpty:
        def __init__(self, *args, **kwargs):
            pass

        num_rows = 0

    class MockReadMetadata:
        def __init__(self, *args, **kwargs):
            pass

        num_rows = 1

    with setup_tempdir(__name__) as tempdir:

        def mock_create_parquet_metadata(parquet_path, nodes):
            metadata_file = os.path.join(tempdir, "_metadata")
            with open(metadata_file, "w") as fd:
                fd.write("")
            return metadata_file

        outfile = os.path.join(tempdir, "test.sonata")

        with open(outfile, "w") as fd:
            fd.write("")

        with patch("glob.glob", MockGlob.glob):
            with patch("pyarrow.parquet.read_metadata", MockReadMetadataEmpty):
                # check that error is raised if empty input file(s)
                with pytest.raises(AssertionError, match="All .parquet files must be non-empty"):
                    converters.parquet_to_sonata(inpath, outfile, nodes, nodefiles)
            with patch("pyarrow.parquet.read_metadata", MockReadMetadata):
                with patch("subprocess.Popen", MockPopen):
                    with patch(
                        "connectome_manipulator.connectome_manipulation.converters.create_parquet_metadata",
                        mock_create_parquet_metadata,
                    ):
                        # monkeypatch.setattr(test_module, 'create_parquet_metadata',
                        #                    mock_create_parquet_metadata)

                        # check that error is raised if file does not exist after
                        with pytest.raises(AssertionError, match="SONATA"):
                            converters.parquet_to_sonata(inpath, outfile, nodes, nodefiles)

                        # check that the file was removed
                        assert not os.path.exists(outfile)


def test_create_new_file_from_template():
    with setup_tempdir(__name__) as tempdir:
        template = os.path.join(tempdir, "template")
        new_file = os.path.join(tempdir, "new_file")
        orig_text = " # comment here\nnormal line here"
        replacements = {"here": "somewhere", "normal": "cool"}

        with open(template, "w") as fd:
            fd.write(orig_text)

        def read_new_file():
            with open(new_file) as fd:
                return fd.read()

        test_module.create_new_file_from_template(
            new_file, template, replacements, skip_comments=True
        )
        content = read_new_file()

        assert content.find("comment here") > -1
        assert content.find("cool line somewhere") > -1

        test_module.create_new_file_from_template(
            new_file, template, replacements, skip_comments=False
        )
        content = read_new_file()

        assert content.find("comment somewhere") > -1


def test_create_sonata_config():
    def json_read(path):
        with open(path, "r") as fd:
            return json.load(fd)

    config_path = os.path.join(TEST_DATA_DIR, "circuit_sonata.json")
    config = Circuit(config_path).config

    orig_edges = "edges.h5"
    new_edges_name = "edges-test.h5"
    population_name = "my_edges"

    with setup_tempdir(__name__) as tempdir:
        new_config_path = os.path.join(tempdir, "circuit_sonata.json")
        new_edges_path = os.path.join(tempdir, new_edges_name)

        test_module.create_sonata_config(config, new_config_path, new_edges_path, population_name)
        res = json_read(new_config_path)

        assert os.path.basename(res["networks"]["edges"][0]["edges_file"]) == new_edges_name

        assert res["manifest"] == {"$BASE_DIR": "."}
        assert res["networks"]["nodes"][0]["nodes_file"] == os.path.join(TEST_DATA_DIR, "nodes.h5")
        assert res["networks"]["edges"][0]["edges_file"] == "$BASE_DIR/edges-test.h5"


def test_create_workflow_config():
    mapping = {
        v: f"${v}"
        for v in [
            "CIRCUIT_NAME",
            "CIRCUIT_DESCRIPTION",
            "CIRCUIT_TYPE",
            "CIRCUIT_CONFIG",
            "DATE",
            "FILE_NAME",
        ]
    }

    with setup_tempdir(__name__) as tempdir:
        template_file = os.path.join(tempdir, "template.json")
        with open(template_file, "w") as fd:
            json.dump(mapping, fd)

        circuit_path = os.path.join(tempdir, "fake_circuit.json")
        blue_config = os.path.join(tempdir, "fake_CircuitConfig")
        output_path = os.path.join(tempdir, "fake_outpath")
        workflow_path = os.path.join(output_path, "workflows")

        for manip_name in ["fake_manipulation", None]:
            test_module.create_workflow_config(
                circuit_path, blue_config, manip_name, output_path, template_file
            )

            assert os.path.isdir(workflow_path)

            dirlist = [os.path.join(workflow_path, p) for p in os.listdir(workflow_path)]
            assert len(dirlist) == 1

            with open(dirlist[0], "r") as fd:
                res = json.load(fd)

            os.remove(dirlist[0])

            for k in mapping:
                assert res[k] != mapping[k]



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os

import numpy as np
import pandas as pd
import pytest
import re
from bluepysnap import Circuit
from mock import Mock, patch
from numpy.testing import assert_array_equal
from scipy.sparse import csc_matrix

from utils import TEST_DATA_DIR, setup_tempdir
import connectome_manipulator.model_building.conn_prob_adj as test_module
from connectome_manipulator.access_functions import get_connections
from connectome_manipulator.model_building import model_types


def test_extract():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]

    # Case 1: Empty node selection
    with pytest.raises(AssertionError, match=re.escape("Empty src/tgt node selection(s)")):
        test_module.extract(circuit, sel_src=[], sel_dest=[])

    # Case 2: Non-empty nodes selection
    sel_src = {"layer": "LA"}
    sel_tgt = {"layer": "LB"}
    src_ids = nodes[0].ids(sel_src)
    tgt_ids = nodes[1].ids(sel_tgt)

    conns = get_connections(edges, src_ids, tgt_ids)
    adj_mat = np.zeros((len(src_ids), len(tgt_ids)), dtype=bool)
    for _s, _t in conns:
        adj_mat[np.where(src_ids == _s)[0], np.where(tgt_ids == _t)[0]] = True

    res = test_module.extract(circuit, sel_src=sel_src, sel_dest=sel_tgt)
    assert_array_equal(src_ids, res["src_node_ids"])
    assert_array_equal(tgt_ids, res["tgt_node_ids"])
    assert res["adj_mat"].format.lower() == "csc"
    assert res["adj_mat"].dtype == bool
    assert_array_equal(res["adj_mat"].toarray(), adj_mat)


def test_build():
    # Define (random) adjacency matrix
    np.random.seed(0)
    src_node_ids = np.arange(5, 20)
    tgt_node_ids = np.arange(50, 100)
    adj_mat = csc_matrix(np.random.rand(len(src_node_ids), len(tgt_node_ids)) > 0.75)

    # Case 1: Non-inverted model
    inverted = False
    model = test_module.build(adj_mat, src_node_ids, tgt_node_ids, inverted=inverted)
    # Check internal representation
    assert_array_equal(src_node_ids, model.get_src_nids())
    assert_array_equal(tgt_node_ids, model.get_tgt_nids())
    assert_array_equal(adj_mat.toarray(), model.get_adj_matrix().toarray())
    assert not model.is_inverted()
    # Check model output
    assert_array_equal(
        adj_mat.toarray().astype(float), model.apply(src_nid=src_node_ids, tgt_nid=tgt_node_ids)
    )

    # Case 2: Inverted model
    inverted = True
    model = test_module.build(adj_mat, src_node_ids, tgt_node_ids, inverted=inverted)
    # Check internal representation
    assert_array_equal(src_node_ids, model.get_src_nids())
    assert_array_equal(tgt_node_ids, model.get_tgt_nids())
    assert_array_equal(adj_mat.toarray(), model.get_adj_matrix().toarray())
    assert model.is_inverted()
    # Check (inverted) model output
    assert_array_equal(
        adj_mat.toarray().astype(float),
        1.0 - model.apply(src_nid=src_node_ids, tgt_nid=tgt_node_ids),
    )



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Main module for model building:

- main(...): Main function for running model building, given a model building config dict
  * Loads a SONATA connectome
  * Extracts connectivity specific data
  * Fits a model to data
  * Stores the data and model to disk
  * Visualizes and compares data and model
- create_model_config_per_pathway(...): Pathway-specific model building wrapper function
"""

from copy import deepcopy
import importlib
import json
import logging
import os.path
import pickle
import time

from bluepysnap.circuit import Circuit
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from connectome_manipulator import log
from connectome_manipulator.model_building import model_types
from connectome_manipulator.access_functions import get_node_ids, get_edges_population

logger = logging.getLogger(__name__)
plt.set_loglevel("info")  # To get rid of a huge list of Matplotlib's DEBUG output


def create_model_config_per_pathway(
    model_config,
    grouped_by,
    src_sel_key="sel_src",
    dest_sel_key="sel_dest",
    group_fct=None,
    group_type=None,
    edge_popul_name=None,
):
    """Create model config dict for pathways between all pairs of groups (e.g. layer, mtype, ...)."""
    # Check model config
    assert "model" in model_config.keys(), 'ERROR: "model" key missing in model_config!'
    assert "working_dir" in model_config.keys(), 'ERROR: "working_dir" key missing in model_config!'
    assert "out_dir" in model_config.keys(), 'ERROR: "out_dir" key missing in model_config!'
    assert (
        "name" in model_config["model"].keys()
    ), 'ERROR: "name" key missing in model_config["model"]!'
    assert (
        "fct" in model_config["model"].keys()
    ), 'ERROR: "fct" key missing in model_config["model"]!'
    assert (
        "kwargs" in model_config["model"]["fct"].keys()
    ), 'ERROR: "kwargs" key missing in model_config["model"]["fct"]!'

    if not isinstance(grouped_by, list):
        grouped_by = [grouped_by]
    grouped_by_name = "-".join(grouped_by)

    if group_type is None or (isinstance(group_type, str) and group_type.upper() == "BOTH"):
        src_grouping = True
        tgt_grouping = True
        print("INFO: Using PRE/POST grouping")
    elif isinstance(group_type, str) and group_type.upper() == "PRE":
        src_grouping = True
        tgt_grouping = False
        grouped_by_name = f"PRE-{grouped_by_name}"
        print("INFO: Using PRE grouping only")
    elif isinstance(group_type, str) and group_type.upper() == "POST":
        src_grouping = False
        tgt_grouping = True
        grouped_by_name = f"POST-{grouped_by_name}"
        print("INFO: Using POST grouping only")
    else:
        assert (
            False
        ), f'ERROR: Grouping type "{group_type}" unknown. Must be PRE, POST, or BOTH (=default).!'

    if group_fct is not None:
        if not isinstance(group_fct, list):
            group_fct = [group_fct]
        assert len(group_fct) == len(
            grouped_by
        ), f"ERROR: Group functions to be provided for {len(grouped_by)} groups!"

    # Load circuit
    circuit_config = model_config["circuit_config"]
    circuit = Circuit(circuit_config)
    print(f"INFO: Circuit loaded: {circuit_config}")

    # Select edge population
    edges = get_edges_population(circuit, popul_name=edge_popul_name)

    # Select corresponding source/target nodes populations
    src_nodes = edges.source
    tgt_nodes = edges.target

    # Find pathways between pairs of groups (within current selection)
    if src_grouping:
        sel_src = model_config["model"]["fct"]["kwargs"].get(src_sel_key)
        if sel_src is not None:
            assert isinstance(
                sel_src, dict
            ), "ERROR: Source node selection must be a dict or empty!"  # Otherwise, it cannot be merged with pathway selection

        assert np.all(
            [g in src_nodes.property_names for g in grouped_by]
        ), f'ERROR: "{grouped_by_name}" property not found in source nodes!'
        node_ids_src = get_node_ids(src_nodes, sel_src)
        src_types_nodes = src_nodes.get(node_ids_src, properties=grouped_by)
        src_types = sorted(src_types_nodes.groupby(grouped_by).indices.keys())
        if len(grouped_by) == 1:
            src_types = [[s] for s in src_types]

        if group_fct is not None:
            src_types_nodes_base = pd.concat(
                [src_types_nodes[grouped_by[i]].apply(group_fct[i]) for i in range(len(group_fct))],
                axis=1,
            )
            src_types_base = sorted(src_types_nodes_base.groupby(grouped_by).indices.keys())
            if len(grouped_by) == 1:
                src_types_base = [[s] for s in src_types_base]
            src_types = [
                [
                    list(
                        np.unique(
                            src_types_nodes[grouped_by[i]][
                                src_types_nodes_base[grouped_by[i]] == s[i]
                            ]
                        )
                    )
                    for i in range(len(group_fct))
                ]
                for s in src_types_base
            ]
        else:
            src_types_base = src_types
    else:
        src_types = src_types_base = [None]

    if tgt_grouping:
        sel_dest = model_config["model"]["fct"]["kwargs"].get(dest_sel_key)
        if sel_dest is not None:
            assert isinstance(
                sel_dest, dict
            ), "ERROR: Target node selection must be a dict or empty!"  # Otherwise, it cannot be merged with pathway selection

        assert np.all(
            [g in tgt_nodes.property_names for g in grouped_by]
        ), f'ERROR: "{grouped_by_name}" property not found in target nodes!'
        node_ids_dest = get_node_ids(tgt_nodes, sel_dest)
        tgt_types_nodes = tgt_nodes.get(node_ids_dest, properties=grouped_by)
        tgt_types = sorted(tgt_types_nodes.groupby(grouped_by).indices.keys())
        if len(grouped_by) == 1:
            tgt_types = [[t] for t in tgt_types]

        if group_fct is not None:
            tgt_types_nodes_base = pd.concat(
                [tgt_types_nodes[grouped_by[i]].apply(group_fct[i]) for i in range(len(group_fct))],
                axis=1,
            )
            tgt_types_base = sorted(tgt_types_nodes_base.groupby(grouped_by).indices.keys())
            if len(grouped_by) == 1:
                tgt_types_base = [[t] for t in tgt_types_base]
            tgt_types = [
                [
                    list(
                        np.unique(
                            tgt_types_nodes[grouped_by[i]][
                                tgt_types_nodes_base[grouped_by[i]] == t[i]
                            ]
                        )
                    )
                    for i in range(len(group_fct))
                ]
                for t in tgt_types_base
            ]
        else:
            tgt_types_base = tgt_types
    else:
        tgt_types = tgt_types_base = [None]

    # Create list of model configs per pathway
    model_build_name = model_config["model"]["name"]
    model_config_pathways = []
    for s, sname in zip(src_types, src_types_base):
        for t, tname in zip(tgt_types, tgt_types_base):
            m_dict = deepcopy(model_config)
            if src_grouping:
                if sel_src is None:
                    m_dict["model"]["fct"]["kwargs"].update(
                        {
                            src_sel_key: {
                                k: (v.tolist() if hasattr(v, "tolist") else v)
                                for k, v in zip(grouped_by, s)
                            }
                        }
                    )  # tolist ... to get rid of hidden numpy data types (e.g., numpy.int64)
                else:
                    m_dict["model"]["fct"]["kwargs"][src_sel_key].update(
                        {
                            k: (v.tolist() if hasattr(v, "tolist") else v)
                            for k, v in zip(grouped_by, s)
                        }
                    )  # tolist ... to get rid of hidden numpy data types (e.g., numpy.int64)
                pre_str = ["-".join([str(sn) for sn in sname]).replace(":", "_")]
            else:
                pre_str = []
            if tgt_grouping:
                if sel_dest is None:
                    m_dict["model"]["fct"]["kwargs"].update(
                        {
                            dest_sel_key: {
                                k: (v.tolist() if hasattr(v, "tolist") else v)
                                for k, v in zip(grouped_by, t)
                            }
                        }
                    )  # tolist ... to get rid of hidden numpy data types (e.g., numpy.int64)
                else:
                    m_dict["model"]["fct"]["kwargs"][dest_sel_key].update(
                        {
                            k: (v.tolist() if hasattr(v, "tolist") else v)
                            for k, v in zip(grouped_by, t)
                        }
                    )  # tolist ... to get rid of hidden numpy data types (e.g., numpy.int64)
                post_str = ["-".join([str(tn) for tn in tname]).replace(":", "_")]
            else:
                post_str = []
            m_dict["model"]["name"] += f'__{grouped_by_name}__{"-".join(pre_str + post_str)}'
            m_dict["working_dir"] = os.path.join(
                m_dict["working_dir"], model_build_name + f"__{grouped_by_name}_pathways"
            )
            m_dict["out_dir"] = os.path.join(
                m_dict["out_dir"], model_build_name + f"__{grouped_by_name}_pathways"
            )
            model_config_pathways.append(m_dict)

    print(
        f'INFO: Created model configurations for {len(model_config_pathways)} pathways between {len(src_types)}x{len(tgt_types)} {grouped_by_name}{"e" if grouped_by_name[-1] == "s" else ""}s'
    )

    return model_config_pathways


def main(model_config_input, show_fig=False, force_recomp=False, cv_folds=None):  # pragma: no cover
    """Main entry point for connectome model building."""
    # Check model building config(s)
    if not isinstance(model_config_input, list):
        assert isinstance(
            model_config_input, dict
        ), "ERROR: model_config_input must be of type list or dict!"
        model_config_input = [model_config_input]

    if len(model_config_input) > 1:
        print(
            f'INFO: Building {len(model_config_input)} models: {model_config_input[0]["model"]["name"]}..{model_config_input[-1]["model"]["name"]}'
        )

    # Run model building
    for midx, model_config in enumerate(model_config_input):
        if len(model_config_input) > 1:
            print(
                f'\n>>> BUILDING MODEL {midx + 1}/{len(model_config_input)}: {model_config["model"]["name"]} <<<',
                flush=True,
            )

        if np.isscalar(force_recomp):
            force_reextract = force_recomp
            force_rebuild = force_recomp
        else:
            assert len(force_recomp) == 2, 'ERROR: Two "force_recomp" entries expected!'
            force_reextract = force_recomp[0]
            force_rebuild = force_recomp[1]

        # Prepare saving
        model_build_name = model_config["model"]["name"]
        # Where to put output/figures
        out_dir = os.path.join(model_config["out_dir"], "output", model_build_name)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        data_dir = os.path.join(model_config["working_dir"], "data")  # Where to load/put data
        if not os.path.exists(os.path.split(data_dir)[0]):
            os.makedirs(os.path.split(data_dir)[0])

        model_dir = os.path.join(model_config["working_dir"], "model")  # Where to put model
        if not os.path.exists(os.path.split(model_dir)[0]):
            os.makedirs(os.path.split(model_dir)[0])

        # Initialize logger file
        log.create_log_file(out_dir, "model_building")

        # Prepare computation module
        comp_source = model_config["model"]["fct"]["source"]
        comp_kwargs = model_config["model"]["fct"]["kwargs"]

        comp_module = importlib.import_module(
            "connectome_manipulator.model_building." + comp_source
        )
        log.log_assert(
            hasattr(comp_module, "extract"),
            f'ERROR: Model building module "{comp_source}" requires extract() function!',
        )
        log.log_assert(
            hasattr(comp_module, "build"),
            f'ERROR: Model building module "{comp_source}" requires build() function!',
        )
        log.log_assert(
            hasattr(comp_module, "plot"),
            f'ERROR: Model building module "{comp_source}" requires plot() function!',
        )

        # Load circuit
        circuit_config = model_config["circuit_config"]
        circuit = Circuit(circuit_config)
        log.info(f"Circuit loaded: {circuit_config}")

        # Prepare cross-validation (optional)
        if cv_folds is None:
            cv_folds_param = model_config.get("CV_folds")
        else:
            cv_folds_param = cv_folds
            if "CV_folds" in model_config:
                log.debug(
                    f"Overwriting CV_folds ({model_config['CV_folds']}) from configuration file with command line argument --cv-folds={cv_folds}"
                )
        if cv_folds_param is not None and cv_folds_param > 1:
            cv_n = cv_folds_param
            cv_data = ["train", "test"]
        else:
            cv_n = 1
            cv_data = ["all"]

        for cv_i in range(cv_n):
            for cv_dset in cv_data:

                if cv_n > 1:
                    log.info(f'>>> CV {cv_i + 1}/{cv_n} ("{cv_dset}") <<<')

                if cv_dset == "all":
                    cv_dict = None
                    cv_dstr = ""
                    cv_mstr = ""
                else:
                    cv_dict = {
                        "n_folds": cv_n,
                        "fold_idx": cv_i,
                        "training_set": cv_dset == "train",
                    }
                    cv_dstr = f"__CV{cv_n}-{cv_i+1}-{cv_dset}"
                    cv_mstr = f"__CV{cv_n}-{cv_i+1}-train"  # Model is always built on training data

                np.random.seed(model_config.get("seed", 123456))

                # Extract data (or load from file)
                data_file = os.path.join(data_dir, model_build_name + cv_dstr + ".pickle")
                if os.path.exists(data_file) and not force_reextract:
                    # Load from file
                    log.info(f"Loading data from {data_file}")
                    with open(data_file, "rb") as f:
                        data_dict = pickle.load(f)
                else:
                    # Compute & save to file
                    t_start = time.time()
                    data_dict = comp_module.extract(circuit, **comp_kwargs, CV_dict=cv_dict)
                    log.info(f"<TIME ELAPSED (data extraction): {time.time() - t_start:.1f}s>")
                    log.info(f"Writing data to {data_file}")
                    if not os.path.exists(os.path.split(data_file)[0]):
                        os.makedirs(os.path.split(data_file)[0])
                    with open(data_file, "wb") as f:
                        pickle.dump(data_dict, f)

                # Build model (or load from file)
                model_name = model_build_name + cv_mstr
                model_file = os.path.join(model_dir, model_name + ".json")
                if os.path.exists(model_file) and (not force_rebuild or cv_dset == "test"):
                    # Load from file
                    log.info(f"Loading model from {model_file}")
                    model = model_types.AbstractModel.model_from_file(model_file)
                else:
                    # Compute & save to file
                    log.log_assert(
                        cv_dset != "test",
                        "ERROR: CV test data cannot be used for building a model!",
                    )
                    t_start = time.time()
                    model = comp_module.build(**data_dict, **comp_kwargs)
                    log.info(f"<TIME ELAPSED (model building): {time.time() - t_start:.1f}s>")
                    log.info(f"Writing model to {model_file}")
                    if not os.path.exists(model_dir):
                        os.makedirs(model_dir)
                    model.save_model(model_dir, model_name)

                    # Save model config dict for reproducibility
                    with open(os.path.join(out_dir, "model_config.json"), "w") as f:
                        json.dump(model_config, f, indent=2)

                # Visualize data vs. model
                cv_out_dir = os.path.join(out_dir, cv_dstr.strip("_"))
                if not os.path.exists(cv_out_dir):
                    os.makedirs(cv_out_dir)
                comp_module.plot(**data_dict, **comp_kwargs, model=model, out_dir=cv_out_dir)

        if show_fig:
            plt.show()
        else:
            plt.close("all")



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Null manipulation module - does not do anything."""

from connectome_manipulator import log
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation


class NullManipulation(Manipulation):
    """Dummy manipulation class performing no manipulation at all:

    This manipulation is intended as a control condition to run the manipulation
    pipeline without actually manipulating the connectome, and can be applied through
    the :func:`apply` method.
    """

    def apply(self, split_ids, **kwargs):
        """Applies a null manipulation, i.e., the connectome is left unchanged.

        Args:
            split_ids (list-like): List of neuron IDs that are part of the current data split; will be automatically provided by the manipulator framework - Not used
            **kwargs: Additional keyword arguments - Not used
        """
        log.info("Nothing to do")



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import numpy as np
import os
import pandas as pd
import pytest

from bluepysnap import Circuit

from utils import TEST_DATA_DIR
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation
from connectome_manipulator.connectome_manipulation.converters import EdgeWriter
from connectome_manipulator import log


@pytest.fixture
def manipulation():
    m = Manipulation.get("syn_prop_alteration")
    return m


def test_apply(manipulation):
    log.setup_logging()  # To have data logging in a defined state

    c = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = c.edges[c.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)

    # Case 1: Check prop_sel
    ## (a) Invalid selection
    for prop_sel in ["@source_node", "@target_node"]:
        with pytest.raises(AssertionError):
            writer = EdgeWriter(None, existing_edges=edges_table.copy())
            manipulation(nodes, writer).apply(tgt_ids, prop_sel, {})

    ## (b) Valid selection
    pct = 100.0
    new_value = {"mode": "setval", "value": -1.0}
    for prop_sel in ["u_syn", "n_rrp_vesicles", "delay", "afferent_section_pos"]:
        props_nonsel = np.setdiff1d(edges_table.columns, prop_sel)
        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            prop_sel,
            new_value,
            sel_src=None,
            sel_dest=None,
            syn_filter=None,
            amount_pct=pct,
        )
        res = writer.to_pandas()

        assert edges_table[props_nonsel].equals(
            res[props_nonsel]
        ), "ERROR: Non-selected property values changed!"
        assert np.all(
            res[prop_sel] == new_value["value"]
        ), "ERROR: Selected property values not modified correctly!"

    # Case 2: Check pct
    prop_sel = "u_syn"
    props_nonsel = np.setdiff1d(edges_table.columns, prop_sel)
    new_value = {"mode": "setval", "value": -1.0}
    for pct in np.linspace(0, 100, 6):
        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            prop_sel,
            new_value,
            sel_src=None,
            sel_dest=None,
            syn_filter=None,
            amount_pct=pct,
        )
        res = writer.to_pandas()

        assert edges_table[props_nonsel].equals(
            res[props_nonsel]
        ), "ERROR: Non-selected property values changed!"
        assert (
            100.0 * np.sum(res[prop_sel] == new_value["value"]) / res.shape[0] == pct
        ), "ERROR: Wrong percentage of synapses changed to target value!"
        assert (
            100.0 * np.sum(res[prop_sel] == edges_table[prop_sel]) / res.shape[0] == 100.0 - pct
        ), "ERROR: Wrong percentage of synapses unchanged!"

    # Case 3: Check sel_src/sel_dest
    pct = 100.0
    for src_class in ["EXC", "INH"]:
        for tgt_class in ["EXC", "INH"]:
            sel_src = {"synapse_class": src_class}
            sel_dest = {"synapse_class": tgt_class}
            writer = EdgeWriter(None, existing_edges=edges_table.copy())
            manipulation(nodes, writer).apply(
                tgt_ids,
                prop_sel,
                new_value,
                sel_src=sel_src,
                sel_dest=sel_dest,
                syn_filter=None,
                amount_pct=pct,
            )
            res = writer.to_pandas()

            assert edges_table[props_nonsel].equals(
                res[props_nonsel]
            ), "ERROR: Non-selected property values changed!"
            sel_idx = np.logical_and(
                np.isin(edges_table["@source_node"], nodes[0].ids(sel_src)),
                np.isin(edges_table["@target_node"], nodes[1].ids(sel_dest)),
            )
            assert np.all(
                res[prop_sel][sel_idx] == new_value["value"]
            ), "ERROR: Selected property values not modified correctly!"
            assert np.all(
                res[prop_sel][~sel_idx] == edges_table[prop_sel][~sel_idx]
            ), "ERROR: Non-selected property values changed!"

    # Case 4: Check syn_filter
    filt_prop = "n_rrp_vesicles"
    filt_val = [2, 3]
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids,
        prop_sel,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter={filt_prop: filt_val},
        amount_pct=pct,
    )
    res = writer.to_pandas()

    assert edges_table[props_nonsel].equals(
        res[props_nonsel]
    ), "ERROR: Non-selected property values changed!"
    sel_idx = np.isin(edges_table[filt_prop], filt_val)
    assert np.all(
        res[prop_sel][sel_idx] == new_value["value"]
    ), "ERROR: Selected property values not modified correctly!"
    assert np.all(
        res[prop_sel][~sel_idx] == edges_table[prop_sel][~sel_idx]
    ), "ERROR: Non-selected property values changed!"

    # Case 5: Check modes
    ## (a) Constant value
    new_value = {"mode": "setval", "value": -1.0}
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids,
        prop_sel,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter=None,
        amount_pct=pct,
    )
    res = writer.to_pandas()

    assert edges_table[props_nonsel].equals(
        res[props_nonsel]
    ), "ERROR: Non-selected property values changed!"
    assert np.all(
        res[prop_sel] == new_value["value"]
    ), "ERROR: Selected property values not modified correctly!"

    ## (b) Scaling factor
    new_value = {"mode": "scale", "factor": -1.0}
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids,
        prop_sel,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter=None,
        amount_pct=pct,
    )
    res = writer.to_pandas()

    assert edges_table[props_nonsel].equals(
        res[props_nonsel]
    ), "ERROR: Non-selected property values changed!"
    assert np.all(
        res[prop_sel] == edges_table[prop_sel] * new_value["factor"]
    ), "ERROR: Selected property values not modified correctly!"

    ## (c) Offset value
    new_value = {"mode": "offset", "value": 1.234}
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids,
        prop_sel,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter=None,
        amount_pct=pct,
    )
    res = writer.to_pandas()

    assert edges_table[props_nonsel].equals(
        res[props_nonsel]
    ), "ERROR: Non-selected property values changed!"
    assert np.all(
        res[prop_sel] == edges_table[prop_sel] + new_value["value"]
    ), "ERROR: Selected property values not modified correctly!"

    ## (d) Shuffling across synapses
    new_value = {"mode": "shuffle"}
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids,
        prop_sel,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter=None,
        amount_pct=pct,
    )
    res = writer.to_pandas()

    assert edges_table[props_nonsel].equals(
        res[props_nonsel]
    ), "ERROR: Non-selected property values changed!"
    assert np.all(np.isin(res[prop_sel], edges_table[prop_sel])) and not res[prop_sel].equals(
        edges_table[prop_sel]
    ), "ERROR: Selected property values not modified correctly!"

    ## (e) Random value (constant)
    new_value = {"mode": "randval", "rng": "normal", "kwargs": {"loc": -1.0, "scale": 0.0}}
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids,
        prop_sel,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter=None,
        amount_pct=pct,
    )
    res = writer.to_pandas()

    assert edges_table[props_nonsel].equals(
        res[props_nonsel]
    ), "ERROR: Non-selected property values changed!"
    assert np.all(
        res[prop_sel] == new_value["kwargs"]["loc"]
    ), "ERROR: Selected property values not modified correctly!"

    ## (f) Random scaling factor (constant)
    new_value = {"mode": "randscale", "rng": "normal", "kwargs": {"loc": -1.0, "scale": 0.0}}
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids,
        prop_sel,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter=None,
        amount_pct=pct,
    )
    res = writer.to_pandas()

    assert edges_table[props_nonsel].equals(
        res[props_nonsel]
    ), "ERROR: Non-selected property values changed!"
    assert np.all(
        res[prop_sel] == edges_table[prop_sel] * new_value["kwargs"]["loc"]
    ), "ERROR: Selected property values not modified correctly!"

    ## (g) Random additive value (constant)
    new_value = {"mode": "randadd", "rng": "normal", "kwargs": {"loc": -1.0, "scale": 0.0}}
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids,
        prop_sel,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter=None,
        amount_pct=pct,
    )
    res = writer.to_pandas()

    assert edges_table[props_nonsel].equals(
        res[props_nonsel]
    ), "ERROR: Non-selected property values changed!"
    assert np.all(
        res[prop_sel] == edges_table[prop_sel] + new_value["kwargs"]["loc"]
    ), "ERROR: Selected property values not modified correctly!"

    # Case 5: Check range
    ## (a) Lower bound
    new_value = {"mode": "scale", "factor": -1.0, "range": [0.0, 1.0]}
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids,
        prop_sel,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter=None,
        amount_pct=pct,
    )
    res = writer.to_pandas()

    assert edges_table[props_nonsel].equals(
        res[props_nonsel]
    ), "ERROR: Non-selected property values changed!"
    assert np.all(
        res[prop_sel] == new_value["range"][0]
    ), "ERROR: Selected property values not modified correctly!"

    ## (b) Upper bound
    new_value = {"mode": "scale", "factor": 100.0, "range": [0.0, 1.0]}
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids,
        prop_sel,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter=None,
        amount_pct=pct,
    )
    res = writer.to_pandas()

    assert edges_table[props_nonsel].equals(
        res[props_nonsel]
    ), "ERROR: Non-selected property values changed!"
    assert np.all(
        res[prop_sel] == new_value["range"][1]
    ), "ERROR: Selected property values not modified correctly!"



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Connection removal module."""

import os

import numpy as np
import scipy.sparse as sps

from connectome_manipulator import log, profiler
from connectome_manipulator.access_functions import get_node_ids
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation


class ConnectomeRemoval(Manipulation):
    """Connectome manipulation class for removing connections:

    Removes a percentage of randomly selected connections (i.e., all synapses per
    connection) according to certain cell and #synapses/connection selection criteria.
    The manipulation can be applied through the :func:`apply` method.

    Optionally, a connection mask can be provided, in which case only connections within
    that mask will be considered for removal (in addition to the other selecion criteria).
    """

    @profiler.profileit(name="conn_removal")
    def apply(
        self,
        split_ids,
        sel_src=None,
        sel_dest=None,
        amount_pct=100.0,
        min_syn_per_conn=None,
        max_syn_per_conn=None,
        conn_mask_file=None,
        **kwargs,
    ):
        """Applies a removal of randomly selected connections according to certain selection criteria.

        Args:
            split_ids (list-like): List of neuron IDs that are part of the current data split; will be automatically provided by the manipulator framework
            sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
            sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
            amount_pct (float): Percentage of randomly sampled connections to be removed
            min_syn_per_conn (int): Minimum #synapses/connection for connections to be considered for removal
            max_syn_per_conn (int): Maximum #synapses/connection for connections to be considered for removal
            conn_mask_file (str): Optional connection mask file (.npz) containing a sparse adjacency matrix in scipy.sparse.csc_matrix format, exactly matching the size of the selected source/target neuron selections and indexed in increasing order; only connections within that mask will be considered for removal
            **kwargs: Additional keyword arguments - Not used

        Note:
            Input/output edges (synapse) tables are accessed through the ``writer`` object:

            * Loading input edges: ``edges_table = self.writer.to_pandas()``
            * Writing output edges: ``self.writer.from_pandas(edges_table_manip)``
        """
        # pylint: disable=arguments-differ
        log.log_assert(0.0 <= amount_pct <= 100.0, "amount_pct out of range!")

        gids_src = get_node_ids(self.nodes[0], sel_src)
        gids_dest = get_node_ids(self.nodes[1], sel_dest)

        edges_table = self.writer.to_pandas()
        syn_sel_idx = np.logical_and(
            np.isin(edges_table["@source_node"], gids_src),
            np.isin(edges_table["@target_node"], gids_dest),
        )  # All potential synapses to be removed
        conns, syn_conn_idx, num_syn_per_conn = np.unique(
            edges_table[syn_sel_idx][["@source_node", "@target_node"]],
            axis=0,
            return_inverse=True,
            return_counts=True,
        )
        conn_sel = np.ones(conns.shape[0]).astype(bool)

        # Connection mask (optional)
        if conn_mask_file is not None:
            log.log_assert(
                os.path.splitext(conn_mask_file)[-1].lower() == ".npz",
                f'Connection mask file "{conn_mask_file}" not in .npz format!',
            )
            log.log_assert(
                os.path.exists(conn_mask_file),
                f'Connection mask file "{conn_mask_file}" not found!',
            )
            conn_mask = sps.load_npz(conn_mask_file)
            log.log_assert(
                conn_mask.shape[0] == len(gids_src) and conn_mask.shape[1] == len(gids_dest),
                f"Size of connection mask does not match selected number of pre/post neurons (must be <{len(gids_src)}x{len(gids_dest)}>)!",
            )
            log.info(
                f"Loaded <{conn_mask.shape[0]}x{conn_mask.shape[1]}> connection mask with {conn_mask.count_nonzero()} entries"
            )

            # Create index table for converting neuron IDs to matrix indices
            src_conv = sps.csr_matrix(
                (
                    np.arange(len(gids_src), dtype=int),
                    (np.zeros(len(gids_src), dtype=int), gids_src),
                )
            )
            dest_conv = sps.csr_matrix(
                (
                    np.arange(len(gids_dest), dtype=int),
                    (np.zeros(len(gids_dest), dtype=int), gids_dest),
                )
            )
            conns_reindex = np.array(
                [
                    src_conv[0, conns[:, 0]].toarray().flatten(),
                    dest_conv[0, conns[:, 1]].toarray().flatten(),
                ]
            ).T

            # Apply mask
            conn_sel = np.logical_and(
                conn_sel, np.array(conn_mask[conns_reindex[:, 0], conns_reindex[:, 1]]).flatten()
            )

        # Apply syn/conn filters (optional)
        if min_syn_per_conn is not None:
            log.log_assert(min_syn_per_conn >= 1, "min_syn_per_conn out of range!")
            conn_sel = np.logical_and(conn_sel, num_syn_per_conn >= min_syn_per_conn)

        if max_syn_per_conn is not None:
            log.log_assert(max_syn_per_conn >= 1, "max_syn_per_conn out of range!")
            conn_sel = np.logical_and(conn_sel, num_syn_per_conn <= max_syn_per_conn)

        conn_sel_idx = np.where(conn_sel)[0]
        num_conn = len(conn_sel_idx)
        if num_conn == 0:
            log.debug("Selection empty, nothing to remove!")
        num_remove = np.round(amount_pct * num_conn / 100).astype(int)
        conn_idx_remove = np.random.choice(conn_sel_idx, num_remove, replace=False)
        syn_idx_remove = np.isin(syn_conn_idx, conn_idx_remove)

        if min_syn_per_conn is not None and max_syn_per_conn is not None:
            if min_syn_per_conn == max_syn_per_conn:
                syn_per_conn_info = f"with {min_syn_per_conn} syns/conn "
            else:
                syn_per_conn_info = f"with {min_syn_per_conn}-{max_syn_per_conn} syns/conn "
        elif min_syn_per_conn is None and max_syn_per_conn is not None:
            syn_per_conn_info = f"with max {max_syn_per_conn} syns/conn "
        elif min_syn_per_conn is not None and max_syn_per_conn is None:
            syn_per_conn_info = f"with min {min_syn_per_conn} syns/conn "
        else:
            syn_per_conn_info = ""
        log.info(
            f"Removing {num_remove} ({amount_pct}%) of {num_conn} connections {syn_per_conn_info}(sel_src={sel_src}, sel_dest={sel_dest}, {np.sum(syn_idx_remove)} synapses)"
        )

        syn_sel_idx[syn_sel_idx] = syn_idx_remove  # Set actual indices of connections to be removed
        edges_table_manip = edges_table[~syn_sel_idx].copy()

        self.writer.from_pandas(edges_table_manip)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os
import subprocess


def test_parquet2hdf5():
    """Test if parquet2hdf5 function is available."""
    try:
        with subprocess.Popen(
            ["parquet2hdf5", "--version"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env={"PATH": os.getenv("PATH", "")},
        ) as proc:
            print("parquet2hdf5 v" + proc.communicate()[0].decode())
    except FileNotFoundError as e:
        assert False, str(e)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Useful converters for circuits data"""

import glob
import os
import subprocess
from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq

from .. import log, profiler, __version__


_WRITE_THRESHOLD = 128 * 1024**2

_SYNAPSE_PROPERTIES = [
    "source_node_id",
    "target_node_id",
    "afferent_section_id",
    "afferent_section_pos",
    "afferent_section_type",
    "afferent_center_x",
    "afferent_center_y",
    "afferent_center_z",
    "syn_type_id",
    "edge_type_id",
    "delay",
]
_PROPERTY_TYPES = {
    "source_node_id": "int64",
    "target_node_id": "int64",
    "afferent_section_id": "int64",
    "afferent_section_pos": "float32",
    "afferent_section_type": "int64",
    "afferent_center_x": "float32",
    "afferent_center_y": "float32",
    "afferent_center_z": "float32",
    "syn_type_id": "int64",
    "edge_type_id": "int64",
    "delay": "float32",
}
_ID_COLUMN_MAP = {"@target_node": "target_node_id", "@source_node": "source_node_id"}


class EdgeWriter:
    """Writes data to a Parquet file continuously.

    Pandas has facilities to write to Parquet, but the functionality is limited and does
    not allow to append to the Parquet file.

    This class will use the underlying PyArrow methods to open a Parquet file when needed,
    and repeatedly writing more data into the file as required.  Compared to storing a
    growing Pandas DataFrame and dumping it to disk at once, this will allow us to reduce
    our memory consumption.
    """

    def __init__(self, output_file: Path, existing_edges=None, with_delay: bool = True):
        """Initializes the writer

        If `existing_edges` is a Pandas DataFrame, it will be used to pre-populate the
        write buffer.  When appending to the writer, columns may be removed from the
        pre-existing data.

        A default schema is constructed, taking the `with_delay` parameter into account to
        include or exclude a delay column.  This default schema will be used for appending
        data to the writer if there is a column mismatch with any pre-existing data.

        If `output_file` is `None`, no writing of data to disk will happen.  Data will
        have to be retrieved with the `to_pandas` method.
        """
        self._batches = []
        self._batches_size = 0
        self._total_edges = 0
        self._path = output_file
        self._with_delay = with_delay
        self._drop_edge_type_column = False
        self._schema = None
        self._writer = None

        if existing_edges is None:
            self._schema = self._build_schema(with_delay)
        else:
            self.from_pandas(existing_edges)

    def __len__(self):
        """Returns the total number of rows written and cached"""
        return self._total_edges + sum(map(len, self._batches))

    def clear(self):
        """Reset the saved edges in the buffer"""
        self._batches = []

    def __enter__(self):
        """Entry point for context management"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit point for context management"""
        self.close()
        return False

    @staticmethod
    def _build_schema(with_delay):
        return pa.schema(
            [
                pa.field(n, pa.from_numpy_dtype(_PROPERTY_TYPES[n]))
                for n in _SYNAPSE_PROPERTIES
                if with_delay or n != "delay"
            ]
        )

    def from_pandas(self, df):
        """Replace any stored edges with the Pandas DataFrame given

        If a schema has already been set, it will be used to convert the given Pandas
        DataFrame to the internal representation. If no schema has been set, it will be
        derived from the DataFrame.
        """
        self._drop_edge_type_column = "edge_type_id" not in df.columns
        if self._drop_edge_type_column:
            df["edge_type_id"] = 0
        self._batches = [
            pa.record_batch(
                df.rename(columns=_ID_COLUMN_MAP).reset_index(drop=True),
                schema=self._schema,
            )
        ]
        if not self._schema:
            self._schema = self._batches[0].schema

    def append(self, **kwargs):
        """Append new data to the buffer.

        Requires that `kwargs` contains all required data from the default schema or
        pre-existing data. If the pre-existing data has less columns than the passed in
        data, additional columns will be removed from the buffered data.

        Once the internal buffer exceeds the write threshold, data
        will be appended to the associated parquet file.
        """
        missing = {f.name for f in self._schema} - set(kwargs.keys())
        if missing:
            log.warning(
                "Dropping columns '%s' and reverting to default schema", "', '".join(missing)
            )
            self._schema = self._build_schema(self._with_delay)
            columns = [f.name for f in self._schema]
            self._batches = [b.select(columns) for b in self._batches]

        data = [kwargs.pop(f.name) for f in self._schema]
        log.log_assert(not kwargs, "Additional data for edge writing")
        self._batches.append(pa.record_batch(data, schema=self._schema))
        self._batches_size += self._batches[-1].nbytes
        if self._path and self._batches_size > _WRITE_THRESHOLD:
            self._write_batches()

    def _write_batches(self):
        """Write buffers to parquet"""
        with profiler.profileit(name="write_to_parquet"):
            table = pa.Table.from_batches(self._batches)
            self._batches = []
            self._batches_size = 0
            self._total_edges += len(table)
            log.debug("Writing %d edges to disk, total of %d edges", len(table), self._total_edges)
            if len(table) > 0:  # Only write file if any edges exist
                if not self._writer:
                    log.log_assert(not self._path.exists(), "Can't append to open file")
                    log.debug("Opening %s to write", self._path)
                    self._writer = pq.ParquetWriter(self._path, table.schema)
                self._writer.write_table(table)

    def to_pandas(self):
        """Return the buffer as a Pandas DataFrame"""
        df = (
            pa.Table.from_batches(self._batches, schema=self._schema)
            .to_pandas()
            .rename(columns={"target_node_id": "@target_node", "source_node_id": "@source_node"})
        )
        if self._drop_edge_type_column:
            del df["edge_type_id"]
        return df

    def close(self):
        """Close any open Parquet files"""
        if self._path and self._batches:
            self._write_batches()
        if self._writer:
            self._writer.close()
            self._writer = None


def create_parquet_metadata(parquet_path, nodes):
    """Adding metadata file required for parquet conversion using parquet-converters/0.8.0

    [Modified from: https://bbpgitlab.epfl.ch/hpc/circuit-building/spykfunc/-/blob/main/src/spykfunc/functionalizer.py#L328-354]
    """
    parquet_file_list = glob.glob(os.path.join(parquet_path, "*.parquet"))
    if len(parquet_file_list) == 0:
        log.warning("No .parquet files exist, no metadata created!")
        return None  # No metadata if no .parquet files exist

    schema = pq.ParquetDataset(parquet_path, use_legacy_dataset=False).schema
    if schema.metadata:
        metadata = {k.decode(): v.decode() for k, v in schema.metadata.items()}
    else:
        metadata = {}
    metadata.update(
        {
            "source_population_name": nodes[0].name,
            "source_population_size": str(nodes[0].size),
            "target_population_name": nodes[1].name,
            "target_population_size": str(nodes[1].size),
            "version": __version__,
        }
    )  # [Will be reflected as "version" attribute under edges/<population_name> in the resulting SONATA file]
    new_schema = schema.with_metadata(metadata)
    metadata_file = os.path.join(parquet_path, "_metadata")
    pq.write_metadata(new_schema, metadata_file)

    return metadata_file


def parquet_to_sonata(
    input_path, output_file, nodes, done_file, population_name="default", keep_parquet=False
):
    """Convert all parquet file(s) from input path to SONATA format (using parquet-converters tool; recomputes indices!!).

    [IMPORTANT: .parquet to SONATA converter requires same column data types in all files!!
                Otherwise, value over-/underflows may occur due to wrong interpretation of numbers!!]
    [IMPORTANT: All .parquet files used for conversion must be non-empty!!
                Otherwise, value errors (zeros) may occur in resulting SONATA file!!]
    """
    # Check if .parquet files exist and are non-empty [Otherwise, value errors (zeros) may occur in resulting SONATA file]
    input_file_list = glob.glob(os.path.join(input_path, "*.parquet"))
    log.log_assert(len(input_file_list) > 0, "No .parquet files to convert!")
    log.log_assert(
        np.all([pq.read_metadata(f).num_rows > 0 for f in input_file_list]),
        "All .parquet files must be non-empty to be converted to SONATA!",
    )
    log.info(f"Converting {len(input_file_list)} (non-empty) .parquet file(s) to SONATA")

    # Creating metadata file [Required by parquet-converters/0.8.0]
    metadata_file = create_parquet_metadata(input_path, nodes)

    # Running parquet conversion [Requires parquet-converters/0.8.0]
    if os.path.exists(output_file):
        os.remove(output_file)
    with subprocess.Popen(
        ["parquet2hdf5", str(input_path), str(output_file), population_name],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        env={"PATH": os.getenv("PATH", "")},
    ) as proc:
        log.debug(proc.communicate()[0].decode())
    log.log_assert(
        os.path.exists(output_file),
        "Parquet conversion error - SONATA file not created successfully!",
    )

    # Delete temporary parquet files (optional)
    if not keep_parquet:
        log.info(
            f'Deleting {len(input_file_list)} temporary .parquet file(s), "{os.path.split(metadata_file)[-1]}" file, and "{os.path.split(done_file)[-1]}"'
        )
        for fn in input_file_list:
            os.remove(fn)
        os.remove(metadata_file)
        os.remove(done_file)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import numpy as np
import pytest
from mock import patch, Mock
from numpy.testing import assert_equal

from utils import create_fake_module
import connectome_manipulator.connectome_comparison.structural_comparator as test_module


def test_compute_results():
    module_name = "fake_compute"
    comp_dict = {
        "fct": {"source": module_name, "kwargs": {}},
        "res_sel": {"fake_key_1": "", "fake_key_2": ""},
    }

    # Create a fake module in which compute is a Mock and returns comp_dict['res_sel']
    full_module_name = f"connectome_manipulator.connectome_comparison.{module_name}"
    code = f'from mock import Mock; compute=Mock(return_value={comp_dict["res_sel"]})'
    fake_module = create_fake_module(full_module_name, code)

    # NOTE by herttuai on 06/10/2021:
    # this raises an exception, since
    #   np.isin(comp_dict['res_sel'], list(res_dict.keys()))
    #    > array(False)
    # but
    #   np.isin(list(comp_dict['res_sel']), list(res_dict))
    #    > array([True, True])
    #
    # I assume the latter is the expected behavior?
    with pytest.raises(AssertionError, match="ERROR: Specified results entry not found!"):
        test_module.compute_results(None, comp_dict)

    # Check that compute gets called
    fake_module.compute.assert_called()


def test_results_diff():
    diff = np.random.randint(1000)

    res_dict1 = {
        "dict_res": {"data": np.array((666, 1000))},
        "data": np.reshape(np.arange(4), (2, 2)),
        "ultimate_answer": 42,
    }

    res_dict2 = {
        "dict_res": {"data": res_dict1["dict_res"]["data"] + diff},
        "data": res_dict1["data"] + diff,
        "ultimate_answer": res_dict1["ultimate_answer"],
    }

    expected_dict = {
        "dict_res": {"data": np.full_like(res_dict1["dict_res"]["data"], diff)},
        "data": np.full_like(res_dict1["data"], diff),
        "ultimate_answer": res_dict1["ultimate_answer"],
    }

    assert_equal(test_module.results_diff(res_dict1, res_dict2), expected_dict)

    with pytest.raises(AssertionError, match="Results type mismatch!"):
        res_dict3 = res_dict2.copy()
        res_dict3["data"] = "will fail"
        test_module.results_diff(res_dict1, res_dict3)

    with pytest.raises(AssertionError, match="Results data shape mismatch!"):
        res_dict3 = res_dict2.copy()
        res_dict3["data"] = np.zeros(2)
        test_module.results_diff(res_dict1, res_dict3)

    with pytest.raises(AssertionError, match="Results inconsistency!"):
        res_dict3 = res_dict2.copy()
        res_dict3["ultimate_answer"] = 0
        test_module.results_diff(res_dict1, res_dict3)


def test_plot_results():
    module_name = "fake_plot"
    comp_dict = {"fct": {"source": module_name, "kwargs": {}}}
    full_module_name = f"connectome_manipulator.connectome_comparison.{module_name}"

    # Create fake module in which plot is an instance of Mock()
    fake_module = create_fake_module(full_module_name, "from mock import Mock; plot=Mock()")

    test_module.plot_results(None, None, {}, comp_dict)

    # Check that plot was called
    fake_module.plot.assert_called()



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
from bluepysnap import Circuit
from utils import TEST_DATA_DIR

import connectome_manipulator.model_building.delay as test_module


def test_extract():
    c = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = c.edges[c.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)

    bin_size_um = 500
    max_range_um = 2500

    # Extract delays in given bins
    src_pos = nodes[0].positions(edges_table["@source_node"].to_numpy()).to_numpy()
    tgt_pos = edges_table[
        ["afferent_center_x", "afferent_center_y", "afferent_center_z"]
    ].to_numpy()
    dist = np.sqrt(np.sum((src_pos - tgt_pos) ** 2, 1))
    bins = np.arange(0, max_range_um + 1, bin_size_um)
    delays = [
        edges_table[np.logical_and(dist >= bins[idx], dist < bins[idx + 1])]["delay"].to_numpy()
        for idx in range(len(bins) - 1)
    ]

    res = test_module.extract(
        c, bin_size_um, max_range_um, sel_src=None, sel_dest=None, sample_size=None
    )

    assert np.sum(res["dist_count"]) == edges_table.shape[0]
    assert_array_equal(res["dist_bins"], bins)
    assert_array_equal(res["dist_count"], [len(d) for d in delays])
    assert_array_equal(res["dist_delays_mean"], [np.mean(d) for d in delays])
    assert_array_equal(res["dist_delays_std"], [np.std(d) for d in delays])
    assert res["dist_delay_min"] == np.min(np.hstack(delays))


def test_build():
    bin_size_um = 500
    max_range_um = 2500
    bins = np.arange(0, max_range_um + 1, bin_size_um)

    # Check delay model
    np.random.rand(0)
    dist = bins[:-1] + 0.5 * bin_size_um  # Bin centers
    d_coef = [0.1, 0.003]
    d_mean = d_coef[0] + d_coef[1] * dist
    d_std = np.random.rand(len(dist))
    d_min = np.random.rand()

    model = test_module.build(bins, d_mean, d_std, d_min, bin_size_um)
    model_dict = model.get_param_dict()

    assert_array_almost_equal(
        (model_dict["delay_mean_coeff_a"], model_dict["delay_mean_coeff_b"]), d_coef
    )
    assert np.isclose(model_dict["delay_std"], np.mean(d_std))
    assert model_dict["delay_min"] == d_min



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os
import numpy as np
import pandas as pd
import pytest
import re
from numpy.testing import assert_array_equal
from scipy.sparse import coo_matrix, csc_matrix

from utils import setup_tempdir, TEST_DATA_DIR
import connectome_manipulator.model_building.model_types as test_module


def test_PathwayModel():
    model_dict = {"model": "ConnProbModel", "order": 1, "coeff_a": 3.14}

    pos = [0.0, 1.0, 2.0]
    spos = np.array([pos, pos, pos])
    tpos = np.array([pos])

    expected_default = np.array([[3.14], [3.14], [3.14]])

    with pytest.raises(ValueError):
        # Not specifying any mapping: can't specify pathway_specs
        new_dict = model_dict | {"pathway_specs": pd.DataFrame({})}
        test_module.AbstractModel.model_from_dict(new_dict)

    # Passing types with a default-only model
    default_model = test_module.AbstractModel.model_from_dict(model_dict)
    assert np.all(
        default_model.apply(src_type=[0, 1, 2], src_pos=spos, tgt_type=[34], tgt_pos=tpos) == 3.14
    )

    # Passing types
    better_model = test_module.AbstractModel.model_from_dict(
        model_dict
        | {
            "pathway_specs": pd.DataFrame(
                {"src_type": ["A"], "tgt_type": ["A"], "connprob_coeff_a": [2.72]}
            ).set_index(["src_type", "tgt_type"]),
            "src_type_map": {"A": 0, "B": 1},
            "tgt_type_map": {"A": 0, "B": 1},
        }
    )
    assert np.all(
        better_model.apply(src_type=["A", "A", "A"], src_pos=spos, tgt_type=["A"], tgt_pos=tpos)
        == 2.72
    )
    assert np.all(
        better_model.apply(src_type=[0, 0, 0], src_pos=spos, tgt_type=[0], tgt_pos=tpos) == 2.72
    )

    assert_array_equal(
        better_model.apply(src_type=["A", "A", "A"], src_pos=spos, tgt_type=["B"], tgt_pos=tpos),
        expected_default,
    )
    assert_array_equal(
        better_model.apply(src_type="A", src_pos=spos, tgt_type=["B"], tgt_pos=tpos),
        expected_default,
    )
    assert_array_equal(
        better_model.apply(src_type=[1, 1, 1], src_pos=spos, tgt_type=[0], tgt_pos=tpos),
        expected_default,
    )
    assert_array_equal(
        better_model.apply(src_type=1, src_pos=spos, tgt_type=["B"], tgt_pos=tpos), expected_default
    )
    assert_array_equal(better_model.apply(src_pos=spos, tgt_pos=tpos), expected_default)

    with pytest.raises(KeyError):
        better_model.apply(src_type=["A", "B", "C"], src_pos=spos, tgt_pos=tpos, tgt_type=["D"])
    with pytest.raises(IndexError):
        better_model.apply(src_type=[123, 456, 789], src_pos=spos, tgt_pos=tpos, tgt_type=[-666])
    with pytest.raises(ValueError):
        better_model.apply(src_type=[1.0, 2.0, 3.0], src_pos=spos, tgt_pos=tpos, tgt_type=[-6.66])


def test_LinDelayModel():
    np.random.seed(0)
    with setup_tempdir(__name__) as tempdir:
        # Check deterministic delay
        d_coef = [0.1, 0.003]
        d_min = 0.2
        model_dict = {
            "model": "LinDelayModel",
            "delay_mean_coeff_a": d_coef[0],
            "delay_mean_coeff_b": d_coef[1],
            "delay_std": 0.0,
            "delay_min": d_min,
        }
        model = test_module.AbstractModel.model_from_dict(model_dict)
        dist = np.arange(0.0, 1000.0, 1.0)
        d = model.apply(distance=dist)
        assert_array_equal(d, np.maximum(d_min, d_coef[0] + d_coef[1] * dist))

        # Check delay statistics
        d_std = 0.5
        model_dict = {
            "model": "LinDelayModel",
            "delay_mean_coeff_a": d_coef[0],
            "delay_mean_coeff_b": d_coef[1],
            "delay_std": d_std,
            "delay_min": d_min,
        }
        model = test_module.AbstractModel.model_from_dict(model_dict)
        dist = 1000
        N = 10000
        d = model.apply(distance=np.full(N, dist))
        assert np.isclose(np.mean(d), d_coef[0] + d_coef[1] * dist, rtol=1e-2)
        assert np.isclose(np.std(d), d_std, rtol=2e-2)

        # Check model saving/loading
        model_name = "LinDelayModel_TEST"
        model.save_model(tempdir, model_name)
        model2 = test_module.AbstractModel.model_from_file(
            os.path.join(tempdir, model_name + ".json")
        )
        assert model.get_param_dict() == model2.get_param_dict()
        assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
        assert np.all(
            [
                np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                for k in model.get_data_dict().keys()
            ]
        )


def test_PosMapModel():
    np.random.seed(0)
    with setup_tempdir(__name__) as tempdir:
        # Check pos table access
        N = 100
        pos_data = np.random.rand(N, 3)
        model_dict = {"model": "PosMapModel"}
        data_dict = {"pos_table": pd.DataFrame(pos_data, columns=["x", "y", "z"])}
        model = test_module.AbstractModel.model_from_dict(model_dict, data_dict)
        pos_mapping = model.apply(gids=np.arange(N))
        assert_array_equal(pos_mapping, pos_data)

        # Check model saving/loading
        model_name = "PosMapModel_TEST"
        model.save_model(tempdir, model_name)
        model2 = test_module.AbstractModel.model_from_file(
            os.path.join(tempdir, model_name + ".json")
        )
        assert model.get_param_dict() == model2.get_param_dict()
        assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
        assert np.all(
            [
                np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                for k in model.get_data_dict().keys()
            ]
        )


def test_ConnPropsModel():
    np.random.seed(0)
    with setup_tempdir(__name__) as tempdir:
        N_syn = 10000
        cond_mean = 5.0
        cond_std = 0.5
        # Check properties within connection
        model_dict = {
            "model": "ConnPropsModel",
            "src_types": ["L4_MC"],
            "tgt_types": ["L4_MC"],
            "prop_stats": {
                "n_syn_per_conn": {
                    "L4_MC": {"L4_MC": {"type": "constant", "mean": N_syn, "dtype": "int"}}
                },
                "conductance": {
                    "L4_MC": {
                        "L4_MC": {
                            "type": "normal",
                            "mean": cond_mean,
                            "std": cond_std,
                            "shared_within": True,
                            "dtype": "float",
                        }
                    }
                },
            },
        }
        model = test_module.AbstractModel.model_from_dict(model_dict)
        props = model.apply(
            src_type="L4_MC", tgt_type="L4_MC"
        )  # Single connection with N_syn synapses
        assert props.shape[0] == N_syn  # Number of synapses per connection
        assert len(np.unique(props["conductance"])) == 1  # No variation within connection

        with pytest.raises(KeyError, match="WRONG_TYPE"):
            props = model.apply(src_type="WRONG_TYPE", tgt_type="WRONG_TYPE")

        # Check with #synapses externally provided (overwriting internal N_syn)
        nsyn_ext = 10
        props = model.apply(
            src_type="L4_MC", tgt_type="L4_MC", n_syn=nsyn_ext
        )  # Single connection with nsyn synapses
        assert props.shape[0] == nsyn_ext  # Number of synapses per connection
        assert len(np.unique(props["conductance"])) == 1  # No variation within connection

        # Check model w/o #syn/conn provided
        model_dict = {
            "model": "ConnPropsModel",
            "src_types": ["L4_MC"],
            "tgt_types": ["L4_MC"],
            "prop_stats": {
                "conductance": {
                    "L4_MC": {
                        "L4_MC": {
                            "type": "normal",
                            "mean": cond_mean,
                            "std": cond_std,
                            "shared_within": True,
                            "dtype": "float",
                        }
                    }
                },
            },
        }
        model = test_module.AbstractModel.model_from_dict(model_dict)
        with pytest.raises(AssertionError, match=re.escape('"n_syn_per_conn" missing')):
            props = model.apply(src_type="L4_MC", tgt_type="L4_MC")
        props = model.apply(src_type="L4_MC", tgt_type="L4_MC", n_syn=N_syn)
        assert props.shape[0] == N_syn  # Number of synapses per connection
        assert len(np.unique(props["conductance"])) == 1  # No variation within connection

        # Check properties across connection
        model_dict = {
            "model": "ConnPropsModel",
            "src_types": ["L4_MC"],
            "tgt_types": ["L4_MC"],
            "prop_stats": {
                "n_syn_per_conn": {
                    "L4_MC": {"L4_MC": {"type": "constant", "mean": N_syn, "dtype": "int"}}
                },
                "conductance": {
                    "L4_MC": {
                        "L4_MC": {
                            "type": "normal",
                            "mean": cond_mean,
                            "std": cond_std,
                            "shared_within": False,
                            "dtype": "float",
                        }
                    }
                },
            },
        }
        model = test_module.AbstractModel.model_from_dict(model_dict)
        props = model.apply(
            src_type="L4_MC", tgt_type="L4_MC"
        )  # Single connection with N_syn synapses
        assert props.shape[0] == N_syn  # Number of synapses per connection
        assert np.isclose(cond_mean, np.mean(props["conductance"]), rtol=1e-2)
        assert np.isclose(cond_std, np.std(props["conductance"]), rtol=2e-2)

        # Check property bounds
        cond_std = 5.0
        cond_range = [cond_mean - 1.0, cond_mean + 1.0]
        model_dict = {
            "model": "ConnPropsModel",
            "src_types": ["L4_MC"],
            "tgt_types": ["L4_MC"],
            "prop_stats": {
                "n_syn_per_conn": {
                    "L4_MC": {"L4_MC": {"type": "constant", "mean": N_syn, "dtype": "int"}}
                },
                "conductance": {
                    "L4_MC": {
                        "L4_MC": {
                            "type": "normal",
                            "mean": cond_mean,
                            "std": cond_std,
                            "shared_within": False,
                            "lower_bound": cond_range[0],
                            "upper_bound": cond_range[1],
                            "dtype": "float",
                        }
                    }
                },
            },
        }
        model = test_module.AbstractModel.model_from_dict(model_dict)
        props = model.apply(
            src_type="L4_MC", tgt_type="L4_MC"
        )  # Single connection with N_syn synapses
        assert np.all(props["conductance"] <= cond_range[1])
        assert np.all(props["conductance"] >= cond_range[0])

        # Check model saving/loading
        model_name = "ConnPropsModel_TEST"
        model.save_model(tempdir, model_name)
        model2 = test_module.AbstractModel.model_from_file(
            os.path.join(tempdir, model_name + ".json")
        )
        assert model.get_param_dict() == model2.get_param_dict()
        assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
        assert np.all(
            [
                np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                for k in model.get_data_dict().keys()
            ]
        )


def test_ConnProb1stOrderModel():
    np.random.seed(0)
    with setup_tempdir(__name__) as tempdir:
        src_pos = np.zeros([100, 3])
        tgt_pos = np.random.rand(1000, 3) * 1000 - 500
        # Check prob model access
        p_ref = 0.01
        model_dict = {"model": "ConnProb1stOrderModel", "p_conn": p_ref}
        model = test_module.AbstractModel.model_from_dict(model_dict)
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert p.shape[0] == src_pos.shape[0] and p.shape[1] == tgt_pos.shape[0]
        assert np.all(p == p_ref)  # Constant conn. prob.

        # Check model saving/loading
        model_name = "ConnProb1stOrderModel_TEST"
        model.save_model(tempdir, model_name)
        model2 = test_module.AbstractModel.model_from_file(
            os.path.join(tempdir, model_name + ".json")
        )
        assert model.get_param_dict() == model2.get_param_dict()
        assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
        assert np.all(
            [
                np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                for k in model.get_data_dict().keys()
            ]
        )


def test_ConnProb2ndOrderExpModel():
    np.random.seed(0)
    with setup_tempdir(__name__) as tempdir:
        src_pos = np.zeros([100, 3])
        tgt_pos = np.random.rand(1000, 3) * 1000 - 500
        # Check prob model access
        scale = 0.1
        exponent = 0.006
        model_dict = {"model": "ConnProb2ndOrderExpModel", "scale": scale, "exponent": exponent}
        model = test_module.AbstractModel.model_from_dict(model_dict)
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        dist = np.sqrt(np.sum(tgt_pos**2, 1))
        assert p.shape[0] == src_pos.shape[0] and p.shape[1] == tgt_pos.shape[0]
        assert np.all(
            p[0, :] == scale * np.exp(-exponent * dist)
        )  # Exponential dist.-dep. conn. prob.

        # Check model saving/loading
        model_name = "ConnProb2ndOrderExpModel_TEST"
        model.save_model(tempdir, model_name)
        model2 = test_module.AbstractModel.model_from_file(
            os.path.join(tempdir, model_name + ".json")
        )
        assert model.get_param_dict() == model2.get_param_dict()
        assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
        assert np.all(
            [
                np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                for k in model.get_data_dict().keys()
            ]
        )


def test_ConnProb3rdOrderExpModel():
    np.random.seed(0)
    with setup_tempdir(__name__) as tempdir:
        src_pos = np.zeros([100, 3])
        tgt_pos = np.random.rand(1000, 3) * 1000 - 500
        # Check prob model access
        scale = [0.09, 0.11]
        exponent = [0.008, 0.004]
        model_dict = {
            "model": "ConnProb3rdOrderExpModel",
            "scale_P": scale[0],
            "scale_N": scale[1],
            "exponent_P": exponent[0],
            "exponent_N": exponent[1],
            "bip_coord": tgt_pos.shape[1] - 1,
        }
        model = test_module.AbstractModel.model_from_dict(model_dict)
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        dist = np.sqrt(np.sum(tgt_pos**2, 1))
        P_sel = tgt_pos[:, -1] > 0
        N_sel = tgt_pos[:, -1] < 0
        assert p.shape[0] == src_pos.shape[0] and p.shape[1] == tgt_pos.shape[0]
        assert np.all(
            p[0, P_sel] == scale[0] * np.exp(-exponent[0] * dist[P_sel])
        )  # Bipolar (pos) exponential dist.-dep. conn. prob.
        assert np.all(
            p[0, N_sel] == scale[1] * np.exp(-exponent[1] * dist[N_sel])
        )  # Bipolar (neg) exponential dist.-dep. conn. prob.

        # Check special case (dist==0.0)
        tgt_pos[:, -1] = 0.0
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        dist = np.sqrt(np.sum(tgt_pos**2, 1))
        assert p.shape[0] == src_pos.shape[0] and p.shape[1] == tgt_pos.shape[0]
        assert np.all(
            p[0, :]
            == 0.5
            * (scale[0] * np.exp(-exponent[0] * dist) + scale[1] * np.exp(-exponent[1] * dist))
        )  # Bipolar (pos/neg average) exponential dist.-dep. conn. prob.

        # Check model saving/loading
        model_name = "ConnProb3rdOrderExpModel_TEST"
        model.save_model(tempdir, model_name)
        model2 = test_module.AbstractModel.model_from_file(
            os.path.join(tempdir, model_name + ".json")
        )
        assert model.get_param_dict() == model2.get_param_dict()
        assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
        assert np.all(
            [
                np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                for k in model.get_data_dict().keys()
            ]
        )


def test_ConnProb4thOrderLinInterpnReducedModel():
    np.random.seed(0)
    with setup_tempdir(__name__) as tempdir:
        dr_pos = np.arange(0.5, 1.6, 1.0)  # Position at bin centers
        dz_pos = np.arange(0.5, 2.6, 1.0)  # Position at bin centers
        p_ref = np.random.rand(len(dr_pos), len(dz_pos))
        p_table = pd.DataFrame(
            p_ref.flatten(),
            index=pd.MultiIndex.from_product([dr_pos, dz_pos], names=["dr", "dz"]),
            columns=["p"],
        )
        model = test_module.AbstractModel.model_from_dict(
            {"model": "ConnProb4thOrderLinInterpnReducedModel", "axial_coord": 2},
            {"p_conn_table": p_table},
        )

        # Check prob model access (at center positions)
        src_pos = np.zeros([1, 3])
        tgt_pos = np.array(
            [
                p_table.index.get_level_values(0).to_numpy(),
                np.zeros(p_table.shape[0]),
                p_table.index.get_level_values(1).to_numpy(),
            ]
        ).T
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(p_ref.flatten(), p))

        # Check prob model access (interpolation)
        src_pos = np.zeros([1, 3])
        tgt_pos = np.array([[np.mean(dr_pos[:2]), 0.0, dz_pos[0]]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(np.mean(p_ref[:2, 0]), p))
        tgt_pos = np.array([[dr_pos[0], 0.0, np.mean(dz_pos[:2])]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(np.mean(p_ref[0, :2]), p))
        tgt_pos = np.array([[np.mean(dr_pos[:2]), 0.0, np.mean(dz_pos[:2])]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(np.mean(p_ref[:2, :2]), p))

        # Check prob model access (extrapolation)
        src_pos = np.zeros([1, 3])
        tgt_pos = np.array([[0.0, 0.0, dz_pos[0]]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(p_ref[0, 0], p))  # Symmetric towards dr=0
        tgt_pos = np.array([[max(dr_pos) + 0.1, 0.0, dz_pos[0]]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(0.0, p))  # Zero outside range of data positions
        tgt_pos = np.array([[dr_pos[0], 0.0, min(dz_pos) - 0.1]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(0.0, p))  # Zero outside range of data positions
        tgt_pos = np.array([[dr_pos[0], 0.0, max(dz_pos) + 0.1]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(0.0, p))  # Zero outside range of data positions

        # Check model saving/loading
        model_name = "ConnProb4thOrderLinInterpnReducedModel_TEST"
        model.save_model(tempdir, model_name)
        model2 = test_module.AbstractModel.model_from_file(
            os.path.join(tempdir, model_name + ".json")
        )
        assert model.get_param_dict() == model2.get_param_dict()
        assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
        assert np.all(
            [
                np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                for k in model.get_data_dict().keys()
            ]
        )


def test_ConnProb5thOrderLinInterpnReducedModel():
    np.random.seed(0)
    with setup_tempdir(__name__) as tempdir:
        z_pos = np.arange(0.5, 3.6, 1.0)  # Position at bin centers
        dr_pos = np.arange(0.5, 1.6, 1.0)  # Position at bin centers
        dz_pos = np.arange(0.5, 2.6, 1.0)  # Position at bin centers
        p_ref = np.random.rand(len(z_pos), len(dr_pos), len(dz_pos))
        p_table = pd.DataFrame(
            p_ref.flatten(),
            index=pd.MultiIndex.from_product([z_pos, dr_pos, dz_pos], names=["z", "dr", "dz"]),
            columns=["p"],
        )
        model = test_module.AbstractModel.model_from_dict(
            {"model": "ConnProb5thOrderLinInterpnReducedModel", "axial_coord": 2},
            {"p_conn_table": p_table},
        )

        # Check prob model access (at center positions)
        for z_idx in range(len(z_pos)):
            z = z_pos[z_idx]
            src_pos = np.array([[0.0, 0.0, z]])
            tgt_pos = np.array(
                [
                    p_table[p_table.index.get_level_values(0) == z]
                    .index.get_level_values(1)
                    .to_numpy(),
                    np.zeros(np.sum(p_table.index.get_level_values(0) == z)),
                    z
                    + p_table[p_table.index.get_level_values(0) == z]
                    .index.get_level_values(2)
                    .to_numpy(),
                ]
            ).T
            p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
            assert np.all(np.isclose(p_ref[z_idx, :, :].flatten(), p))

        # Check prob model access (interpolation)
        for z_idx in range(len(z_pos)):
            z = z_pos[z_idx]
            src_pos = np.array([[0.0, 0.0, z]])
            tgt_pos = np.array([[np.mean(dr_pos[:2]), 0.0, z + dz_pos[0]]])
            p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
            assert np.all(np.isclose(np.mean(p_ref[z_idx, :2, 0]), p))
            tgt_pos = np.array([[dr_pos[0], 0.0, z + np.mean(dz_pos[:2])]])
            p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
            assert np.all(np.isclose(np.mean(p_ref[z_idx, 0, :2]), p))
            tgt_pos = np.array([[np.mean(dr_pos[:2]), 0.0, z + np.mean(dz_pos[:2])]])
            p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
            assert np.all(np.isclose(np.mean(p_ref[z_idx, :2, :2]), p))
        src_pos = np.array([[0.0, 0.0, np.mean(z_pos[:2])]])
        tgt_pos = np.array([[dr_pos[0], 0.0, np.mean(z_pos[:2]) + dz_pos[0]]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(np.mean(p_ref[:2, 0, 0]), p))
        tgt_pos = np.array([[np.mean(dr_pos[:2]), 0.0, np.mean(z_pos[:2]) + np.mean(dz_pos[:2])]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(np.mean(p_ref[:2, :2, :2]), p))

        # Check prob model access (extrapolation)
        src_pos = np.array([[0.0, 0.0, z_pos[0]]])
        tgt_pos = np.array([[0.0, 0.0, z_pos[0] + dz_pos[0]]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(p_ref[0, 0, 0], p))  # Symmetric towards dr=0
        tgt_pos = np.array([[max(dr_pos) + 0.1, 0.0, z_pos[0] + dz_pos[0]]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(0.0, p))  # Zero outside range of data positions
        tgt_pos = np.array([[dr_pos[0], 0.0, z_pos[0] + min(dz_pos) - 0.1]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(0.0, p))  # Zero outside range of data positions
        tgt_pos = np.array([[dr_pos[0], 0.0, z_pos[0] + max(dz_pos) + 0.1]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(0.0, p))  # Zero outside range of data positions
        src_pos = np.array([[0.0, 0.0, min(z_pos) - 0.1]])
        tgt_pos = np.array([[dr_pos[0], 0.0, min(z_pos) - 0.1 + dz_pos[0]]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(0.0, p))  # Zero outside range of data positions
        src_pos = np.array([[0.0, 0.0, max(z_pos) + 0.1]])
        tgt_pos = np.array([[dr_pos[0], 0.0, max(z_pos) + 0.1 + dz_pos[0]]])
        p = model.apply(src_pos=src_pos, tgt_pos=tgt_pos)
        assert np.all(np.isclose(0.0, p))  # Zero outside range of data positions

        # Check model saving/loading
        model_name = "ConnProb5thOrderLinInterpnReducedModel_TEST"
        model.save_model(tempdir, model_name)
        model2 = test_module.AbstractModel.model_from_file(
            os.path.join(tempdir, model_name + ".json")
        )
        assert model.get_param_dict() == model2.get_param_dict()
        assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
        assert np.all(
            [
                np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                for k in model.get_data_dict().keys()
            ]
        )


def test_ConnProbAdjModel():
    np.random.seed(999)
    with setup_tempdir(__name__) as tempdir:
        # Define (random) adjacency matrix
        src_node_ids = np.arange(5, 20)
        tgt_node_ids = np.arange(50, 100)
        adj_mat = csc_matrix(np.random.rand(len(src_node_ids), len(tgt_node_ids)) > 0.75)

        # Prepare data frames
        src_nodes_table = pd.DataFrame(src_node_ids, columns=["src_node_ids"])
        tgt_nodes_table = pd.DataFrame(tgt_node_ids, columns=["tgt_node_ids"])
        rows, cols = adj_mat.nonzero()
        adj_table = pd.DataFrame({"row_ind": rows, "col_ind": cols})

        # Test non-inverted & inverted model
        for inverted in [False, True]:
            ## (a) Build
            model_dict = {"model": "ConnProbAdjModel", "inverted": inverted}
            data_dict = {
                "src_nodes_table": src_nodes_table,
                "tgt_nodes_table": tgt_nodes_table,
                "adj_table": adj_table,
            }
            model = test_module.AbstractModel.model_from_dict(model_dict, data_dict)

            ## (b) Access
            assert_array_equal(src_node_ids, model.get_src_nids())
            assert_array_equal(tgt_node_ids, model.get_tgt_nids())
            assert_array_equal(adj_mat.toarray(), model.get_adj_matrix().toarray())
            assert model.is_inverted() == inverted
            if inverted:
                assert_array_equal(
                    adj_mat.toarray().astype(float),
                    1.0 - model.apply(src_nid=src_node_ids, tgt_nid=tgt_node_ids),
                )
            else:
                assert_array_equal(
                    adj_mat.toarray().astype(float),
                    model.apply(src_nid=src_node_ids, tgt_nid=tgt_node_ids),
                )

            ## (c) Load/save
            model_name = "ConnProbAdjModel_TEST"
            model.save_model(tempdir, model_name)
            model2 = test_module.AbstractModel.model_from_file(
                os.path.join(tempdir, model_name + ".json")
            )
            assert model.get_param_dict() == model2.get_param_dict()
            assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
            assert np.all(
                [
                    np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                    for k in model.get_data_dict().keys()
                ]
            )


def test_LookupTableModel():
    np.random.seed(999)
    with setup_tempdir(__name__) as tempdir:
        # Define (random) LUT matrix
        src_node_ids = np.arange(5, 20)
        tgt_node_ids = np.arange(50, 100)
        src_nodes_table = pd.DataFrame(src_node_ids, columns=["src_node_ids"])
        tgt_nodes_table = pd.DataFrame(tgt_node_ids, columns=["tgt_node_ids"])
        mat = np.random.rand(len(src_node_ids), len(tgt_node_ids))
        mat[mat < 0.5] = 0  # Make sparse
        mat_coo = coo_matrix(mat, shape=(len(src_node_ids), len(tgt_node_ids)))
        lookup_table = pd.DataFrame(
            {"row_ind": mat_coo.row, "col_ind": mat_coo.col, "value": mat_coo.data}
        )

        # Init. model
        model = test_module.LookupTableModel(
            src_nodes_table=src_nodes_table,
            tgt_nodes_table=tgt_nodes_table,
            lookup_table=lookup_table,
        )

        # Check model
        assert_array_equal(src_node_ids, model.get_src_nids())
        assert_array_equal(tgt_node_ids, model.get_tgt_nids())
        assert_array_equal(mat, model.lut_mat.todense())
        assert_array_equal(mat, model.apply(src_nid=src_node_ids, tgt_nid=tgt_node_ids))

        ## Check load/save
        model_name = "LookupTableModel_TEST"
        model.save_model(tempdir, model_name)
        model2 = test_module.AbstractModel.model_from_file(
            os.path.join(tempdir, model_name + ".json")
        )
        assert model.get_param_dict() == model2.get_param_dict()
        assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
        assert np.all(
            [
                np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                for k in model.get_data_dict().keys()
            ]
        )

        # Check (static) init from sparse matrix
        with pytest.raises(AssertionError, match=re.escape("Matrix must be in sparse format")):
            model = test_module.LookupTableModel.init_from_sparse_matrix(
                mat, src_node_ids, tgt_node_ids
            )
        model = test_module.LookupTableModel.init_from_sparse_matrix(
            mat_coo.tocsc(), src_node_ids, tgt_node_ids
        )
        assert_array_equal(src_node_ids, model.get_src_nids())
        assert_array_equal(tgt_node_ids, model.get_tgt_nids())
        assert_array_equal(mat, model.lut_mat.todense())
        assert_array_equal(mat, model.apply(src_nid=src_node_ids, tgt_nid=tgt_node_ids))


def test_PropsTableModel():
    np.random.seed(999)
    with setup_tempdir(__name__) as tempdir:
        pnames = ["a", "b", "c", "d", "e"]
        ptab = pd.DataFrame(np.random.rand(20, 5), columns=pnames)
        id_max = ptab.shape[0] >> 2
        id_offs = [5, 10]  # Source/target offsets

        ## Init model
        with pytest.raises(AssertionError, match=re.escape("@source_node column required!")):
            model = test_module.PropsTableModel(props_table=ptab)
        ptab["@source_node"] = np.random.randint(id_max, size=ptab.shape[0]) + id_offs[0]
        with pytest.raises(AssertionError, match=re.escape("@target_node column required!")):
            model = test_module.PropsTableModel(props_table=ptab)
        ptab["@target_node"] = np.random.randint(id_max, size=ptab.shape[0]) + id_offs[1]
        model = test_module.PropsTableModel(props_table=ptab)

        ## Check model
        # (a) Check node IDs
        src_nids = np.unique(ptab["@source_node"])
        tgt_nids = np.unique(ptab["@target_node"])
        assert_array_equal(src_nids, model.get_src_nids())
        assert_array_equal(tgt_nids, model.get_tgt_nids())
        conns, cnts = model.get_src_tgt_counts()
        ref_conns, ref_cnts = np.unique(
            ptab[["@source_node", "@target_node"]], axis=0, return_counts=True
        )
        assert_array_equal(conns, ref_conns)
        assert_array_equal(cnts, ref_cnts)
        assert_array_equal(model.get_property_names(), pnames)

        # (b) Check model access (single node IDs)
        for sid in src_nids:
            for tid in tgt_nids:
                ref_tab = ptab[
                    np.logical_and(ptab["@source_node"] == sid, ptab["@target_node"] == tid)
                ][pnames]
                assert_array_equal(model.apply(src_nid=sid, tgt_nid=tid), ref_tab)

        # (c) Check model access (multiple node IDs)
        for sid in src_nids:
            ref_tab = ptab[
                np.logical_and(ptab["@source_node"] == sid, np.isin(ptab["@target_node"], tgt_nids))
            ][pnames]
            assert_array_equal(model.apply(src_nid=sid, tgt_nid=tgt_nids), ref_tab)
        for tid in tgt_nids:
            ref_tab = ptab[
                np.logical_and(np.isin(ptab["@source_node"], src_nids), ptab["@target_node"] == tid)
            ][pnames]
            assert_array_equal(model.apply(src_nid=src_nids, tgt_nid=tid), ref_tab)
        ref_tab = ptab[pnames]
        assert_array_equal(model.apply(src_nid=src_nids, tgt_nid=tgt_nids), ref_tab)

        # (d) Check model access (single properties)
        for p in pnames:
            assert_array_equal(
                model.apply(src_nid=src_nids, tgt_nid=tgt_nids, prop_names=[p]), ptab[[p]]
            )

        # (e) Check model access (full table, including @source/target_node)
        assert_array_equal(
            model.apply(src_nid=src_nids, tgt_nid=tgt_nids, prop_names=ptab.columns), ptab
        )

        # (f) Check model access (selected number of entries)
        for _n in [-1, 2 * ptab.shape[0]]:  # => Assertion error
            with pytest.raises(
                AssertionError, match=re.escape("Selected number of elements out of range!")
            ):
                model.apply(src_nid=src_nids, tgt_nid=tgt_nids, num_sel=_n)
        assert model.apply(src_nid=src_nids, tgt_nid=tgt_nids, num_sel=0).size == 0
        assert_array_equal(
            model.apply(src_nid=src_nids, tgt_nid=tgt_nids, num_sel=ptab.shape[0]), ptab[pnames]
        )
        for _n in range(1, ptab.shape[0], 5):
            assert_array_equal(
                model.apply(src_nid=src_nids, tgt_nid=tgt_nids, num_sel=_n), ptab[pnames][:_n]
            )

        ## Check load/save
        model_name = "PropsTableModel_TEST"
        model.save_model(tempdir, model_name)
        model2 = test_module.AbstractModel.model_from_file(
            os.path.join(tempdir, model_name + ".json")
        )
        assert model.get_param_dict() == model2.get_param_dict()
        assert sorted(model.get_data_dict().keys()) == sorted(model2.get_data_dict().keys())
        assert np.all(
            [
                np.array_equal(model.get_data_dict()[k], model2.get_data_dict()[k])
                for k in model.get_data_dict().keys()
            ]
        )
